Package: rlmstudio
Title: Access and Control LM Studio
Version: 0.2.2
Authors@R: 
    person("Jeffrey", "Girard", , "me@jmgirard.com",
           role = c("aut", "cph", "cre"),
           comment = c(ORCID = "0000-0002-7359-3746"))
Description: A community-maintained 'R' wrapper for the 'LM Studio' command line
    interface and API. Provides functions to manage the local daemon and server,
    download and load models, and interact with Large Language Models (LLMs).
License: MIT + file LICENSE
URL: https://jmgirard.github.io/rlmstudio/,
        https://github.com/jmgirard/rlmstudio
BugReports: https://github.com/jmgirard/rlmstudio/issues
Depends: R (>= 4.1.0)
Imports: cli, httr2, jsonlite, processx, utils
Suggests: httptest2, knitr, mockery, rmarkdown, testthat (>= 3.0.0),
        withr
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-02 18:24:44 UTC; jmgirard
Author: Jeffrey Girard [aut, cph, cre] (ORCID:
    <https://orcid.org/0000-0002-7359-3746>)
Maintainer: Jeffrey Girard <me@jmgirard.com>
Repository: CRAN
Date/Publication: 2026-05-05 18:30:02 UTC
Built: R 4.4.3; ; 2026-05-05 19:08:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,lms_chat_result)
S3method(print,lms_download_status)
export(check_lms_version)
export(has_lms)
export(install_lmstudio)
export(list_models)
export(lms_chat)
export(lms_chat_batch)
export(lms_chat_native)
export(lms_chat_openai)
export(lms_chat_openresponses)
export(lms_daemon_start)
export(lms_daemon_status)
export(lms_daemon_stop)
export(lms_download)
export(lms_download_status)
export(lms_load)
export(lms_path)
export(lms_score_expected)
export(lms_server_start)
export(lms_server_status)
export(lms_server_stop)
export(lms_unload)
export(lms_unload_all)
export(with_lms_daemon)

<!-- badges: start -->

[![Codecov test
coverage](https://codecov.io/gh/jmgirard/rlmstudio/graph/badge.svg)](https://app.codecov.io/gh/jmgirard/rlmstudio)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
<!-- badges: end -->

# rlmstudio

The `rlmstudio` package provides an R interface to the [LM
Studio](https://lmstudio.ai/) CLI and REST API. It allows you to
download local Large Language Models (LLMs), manage server instances,
load or unload models into memory, and generate text directly from your
R console or scripts.

***Disclaimer:** This is an unofficial, community-maintained R package.
It is not affiliated with, endorsed by, or maintained by the creators of
LM Studio.*

## Installation

``` r
# Install release version from CRAN
install.packages("rlmstudio")

# Install development version from GitHub
remotes::install_github("jmgirard/rlmstudio")
```

## Setup and Prerequisites

This package relies on the LM Studio CLI (version 0.4.0 or higher) and
the new v1 REST API. If you do not have LM Studio installed or need to
update your version, the package provides a convenient setup function.

``` r
library(rlmstudio)
install_lmstudio()
## ℹ Opening the LM Studio download page in your default browser...
## ! Please install or update the software, restart R, and try again.
```

## Quick Start

The standard workflow involves starting the LM Studio server, loading a
model into memory, chatting with it, and then cleaning up your
environment.

For a deeper dive into the package architecture and a detailed
explanation of how to use LM Studio with a visual GUI versus a headless
background daemon, please see the package vignettes.

### 1. Start the Server

Before you can load models or generate text, you need to start the LM
Studio backend server.

``` r
lms_server_start()
## ✔ LM Studio server started successfully on the default port.
```

*(Note for headless environments like remote servers or Docker
containers: You must start the background daemon first using
`lms_daemon_start()` before starting the server).*

### 2. Find and Load a Model

You can see which models you currently have available on your machine
using `list_models()`.

``` r
# Returns a clean data frame of available models
my_models <- list_models()
my_models
##      state  type  display_name                  key  architecture  size_gb
## 1 unloaded   llm  Gemma 3n E4B  google/gemma-3n-e4b       gemma3n     5.46
## 2 unloaded   llm   Gemma 3 12B   google/gemma-3-12b        gemma3     7.51
```

If you do not have a model yet, you can download one using its Hugging
Face repository or LM Studio catalog identifier.

``` r
# Download a lightweight model
job_id <- lms_download("google/gemma-3-1b")
lms_download_status(job_id)
## ── Download Job: "job_02c8a1f86e"
## Status: completed
## Progress: 100% (0.72 GB / 0.72 GB)
```

Once a model is downloaded and available, load it into memory.

``` r
# Load the model
lms_load("google/gemma-3-1b", flash_attention = TRUE)
## ✔ Model "google/gemma-3-1b" loaded and verified. [10.1s]
```

### 3. Chat

Use the `lms_chat()` function for quick and easy interactions.

``` r
response <- lms_chat(
  model = "google/gemma-3-1b",
  input = "What are the capitals of each country in North America?"
)

cat(response)
## Okay, here’s a list of the capitals of each country in North America:
##
## * **Canada:** Ottawa
## * **Mexico:** Mexico City (Federal Capital)
## * **United States of America:** Washington, D.C.
##
## Let me know if you’d like to learn more about any of these countries!
```

If you need access to advanced features like Model Context Protocol
(MCP) integrations, structured tool calling, or granular control over
inference parameters, you can pass additional arguments directly to
`lms_chat()`. Setting `simplify = FALSE` will return the raw, unparsed
API list response.

### 4. Batch Processing

If you have a collection of documents or prompts to process,
`lms_chat_batch()` allows you to iterate over a vector of inputs and
capture the results in a tidy format.

``` r
prompts <- c(
  "Summarize the benefits of local LLMs.",
  "Explain why privacy matters in AI."
)

# Process all prompts and return a data frame
results <- lms_chat_batch(
  model = "google/gemma-3-1b",
  inputs = prompts,
  system_prompt = "Answer in two sentences or less."
)

print(results)
## [1] "Local LLMs offer significant benefits like improved privacy, reduced reliance on cloud services, and increased control over data. They also often demonstrate superior performance within specific domains due to training on localized datasets."
## [2] "Privacy matters in AI because sophisticated algorithms require vast amounts of data, and this data often includes personal information. Without adequate safeguards and regulations, AI systems can violate individuals’ privacy rights and lead to potential misuse of sensitive data."
```

### 5. Clean Up

When you are finished, it is good practice to unload the model from
memory and shut down the local server to free up system resources.

``` r
# Unload the specific model
lms_unload("google/gemma-3-1b")
## ✔ Model "google/gemma-3-1b" unloaded successfully. [382ms]

# Stop the local server
lms_server_stop()
## ✔ LM Studio server stopped successfully.
```

## ----setup--------------------------------------------------------------------
library(rlmstudio)
lms_installed <- has_lms()

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----check-status-------------------------------------------------------------
# Check if LM Studio is available on this system
has_lms()

## ----start-server, eval=lms_installed-----------------------------------------
# Start the local server on the default port
lms_server_start()
#> ✔ LM Studio server started successfully on the default port.

## ----download, eval=lms_installed---------------------------------------------
# Download a model using its identifier
model <- "google/gemma-3-1b"
job_id <- lms_download(model)
#> ✔ Initiating download for model: "google/gemma-3-1b"... [973ms]
#> ✔ Download job started successfully. Job ID: "job_02c8a1f86e"

lms_download_status(job_id)
#> ── Download Job: "job_02c8a1f86e"
#> Status: completed
#> Progress: 100% (0.72 GB / 0.72 GB)

## ----wait, echo=FALSE, eval=lms_installed-------------------------------------
if (job_id != "already_downloaded") {
  repeat {
    res <- lms_download_status(job_id)
    if (res$status %in% c("completed", "failed", "error")) {
      break
    }
    Sys.sleep(3)
  }
  if (res$status != "completed") {
    cli::cli_abort("Model download failed.")
  }
}

## ----load, eval=lms_installed-------------------------------------------------
# Standard load
lms_load(model, flash_attention = TRUE)
#> ✔ Model "google/gemma-3-1b" loaded and verified. [30.8s]

## ----simple-chat, eval=lms_installed------------------------------------------
response <- lms_chat(
  model = model,
  input = "Say hello!",
  system_prompt = "Answer in rhymes."
)

cat(response)
#> A friendly face, so bright and new,
#> Hello there, it’s waiting for you!
#>
#> Let's chat and have a joyful spree,
#> Hello there, happy as can be!

## ----teardown, eval=lms_installed---------------------------------------------
# Unload the model
lms_unload(model)
#> ✔ Model "google/gemma-3-1b" unloaded successfully. [431ms]

# Stop the server
lms_server_stop()
#> ✔ LM Studio server stopped successfully.

## ----setup--------------------------------------------------------------------
library(rlmstudio)
lms_installed <- has_lms()

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----check-status-------------------------------------------------------------
# Verify the CLI is available before proceeding
has_lms()

## ----start-daemon, eval=lms_installed-----------------------------------------
# Start the headless engine in the background
lms_daemon_start()

## ----start-server, eval=lms_installed-----------------------------------------
# Start the local server on the default port
lms_server_start()

## ----download, eval=lms_installed---------------------------------------------
# Download a model using its identifier
job_id <- lms_download("qwen/qwen3-4b-2507")
lms_download_status(job_id)

## ----wait, echo=FALSE, eval=lms_installed-------------------------------------
if (job_id != "already_downloaded") {
  repeat {
    res <- lms_download_status(job_id)
    if (res$status %in% c("completed", "failed", "error")) {
      break
    }
    Sys.sleep(3)
  }
  if (res$status != "completed") {
    cli::cli_abort("Model download failed.")
  }
}

## ----list, eval=lms_installed-------------------------------------------------
# View all downloaded models
models <- list_models()

# Filter for unloaded text models
unloaded_llms <- models |>
  subset(type == "llm" & state == "unloaded")
unloaded_llms

## ----load, eval=lms_installed-------------------------------------------------
# Load the model
lms_load("google/gemma-3-1b", flash_attention = TRUE)

## ----chat, eval=lms_installed-------------------------------------------------
response <- lms_chat(
  model = "google/gemma-3-1b",
  input = "Provide just the str_extract() pattern to match all text after the third comma.",
  system_prompt = "You are an expert R programmer familiar with the tidyverse."
)

cat(response)

## ----teardown, eval=lms_installed---------------------------------------------
# 1. Unload the model from memory
lms_unload("google/gemma-3-1b")

# 2. Stop the API server
lms_server_stop()

# 3. Stop the background daemon
lms_daemon_stop()

## ----with-daemon, eval=lms_installed------------------------------------------
# The daemon will start, the code will run, and the daemon will stop on exit.
results <- with_lms_daemon({
  lms_server_start()
  lms_load("google/gemma-3-1b")

  res <- lms_chat("google/gemma-3-1b", "Is the daemon running?")

  lms_server_stop()
  res
})

#' Download a model via REST API
#'
#' @param model Character. The model to download. Accepts model catalog
#'   identifiers (e.g., "openai/gpt-oss-20b") and exact Hugging Face links.
#' @param quantization Character. Optional. Quantization level of the model to
#'   download (e.g., "Q4_K_M"). Only supported for Hugging Face links.
#' @param host Character. The host address of the local server. Defaults to
#'   "http://localhost:1234".
#' @param ... Additional arguments passed to the request.
#'
#' @seealso [LM Studio Download Model
#'   API](https://lmstudio.ai/docs/developer/rest/download)
#'
#' @return A character string containing the download \code{job_id}, or
#'   \code{"already_downloaded"} if already downloaded.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' # Download a model by its HuggingFace identifier
#' job_id <- lms_download("google/gemma-3-1b")
#'
#' # Download with a specific quantization level
#' lms_download("google/gemma-3-1b", quantization = "4bit")
#' }
lms_download <- function(
  model,
  quantization = NULL,
  host = "http://localhost:1234",
  ...
) {
  if (!is_server_running()) {
    cli::cli_alert_danger(
      "The LM Studio server is not running. Run {.fn lms_server_start} first."
    )
    return(invisible(NULL))
  }

  if (is.null(model) || model == "") {
    cli::cli_abort("You must provide a valid model identifier or URL.")
  }

  body <- list(
    model = model,
    quantization = quantization
  )

  # Remove NULLs and merge any additional arguments from dots
  body <- Filter(Negate(is.null), body)
  body <- utils::modifyList(body, list(...))

  # Capture the step ID so we can manually close it later
  step_id <- rlm_progress_step(
    "Initiating download for model: {.val {model}}..."
  )

  resp <- lms_client(host) |>
    httr2::req_url_path("api/v1/models/download") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  # Explicitly complete the progress step before printing subsequent alerts
  rlm_progress_done(step_id)

  if (httr2::resp_status(resp) == 200) {
    resp_data <- httr2::resp_body_json(resp)

    if (
      !is.null(resp_data$status) && resp_data$status == "already_downloaded"
    ) {
      rlm_alert_success("Model {.val {model}} is already downloaded.")
      return(invisible("already_downloaded"))
    }

    if (!is.null(resp_data$job_id)) {
      rlm_alert_success(
        "Download job started successfully. Job ID: {.val {resp_data$job_id}}"
      )
      return(resp_data$job_id)
    }

    rlm_alert_success("Download request succeeded.")
    return(invisible(TRUE))
  }

  err_msg <- tryCatch(
    {
      err_json <- httr2::resp_body_json(resp)
      if (!is.null(err_json$error$message)) {
        err_json$error$message
      } else if (!is.null(err_json$error)) {
        err_json$error
      } else {
        httr2::resp_body_string(resp)
      }
    },
    error = function(e) httr2::resp_body_string(resp)
  )

  if (err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }

  cli::cli_abort(c("x" = "API Download Failed: {err_msg}"))
}

#' Get the status of a download job
#'
#' @param job_id Character. The unique identifier for the download job.
#' @param host Character. The host address of the local server. Defaults to
#'   "http://localhost:1234".
#'
#' @seealso [LM Studio Download Status
#'   API](https://lmstudio.ai/docs/developer/rest/download-status)
#'
#' @return An object of class \code{lms_download_status} containing the download
#'   status.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' job_id <- lms_download("google/gemma-3-1b")
#' status <- lms_download_status(job_id)
#' print(status)
#' }
lms_download_status <- function(job_id, host = "http://localhost:1234") {
  if (!is_server_running()) {
    cli::cli_alert_danger(
      "The LM Studio server is not running. Run {.fn lms_server_start} first."
    )
    return(invisible(NULL))
  }

  if (identical(job_id, "already_downloaded")) {
    out <- list(
      job_id = "N/A",
      status = "already_downloaded"
    )
    class(out) <- c("lms_download_status", "list")
    return(out)
  }

  if (is.null(job_id) || job_id == "") {
    cli::cli_abort("You must provide a valid job_id.")
  }

  resp <- lms_client(host) |>
    httr2::req_url_path(paste0("api/v1/models/download/status/", job_id)) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    out <- httr2::resp_body_json(resp)
    class(out) <- c("lms_download_status", "list")
    return(out)
  }

  err_msg <- tryCatch(
    {
      err_json <- httr2::resp_body_json(resp)
      if (!is.null(err_json$error$message)) {
        err_json$error$message
      } else if (!is.null(err_json$error)) {
        err_json$error
      } else {
        httr2::resp_body_string(resp)
      }
    },
    error = function(e) httr2::resp_body_string(resp)
  )

  if (err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }

  cli::cli_abort(c("x" = "API Status Request Failed: {err_msg}"))
}

#' Print method for LM Studio download status
#'
#' @param x An object of class \code{lms_download_status}.
#' @param ... Additional arguments passed to print.
#'
#' @keywords internal
#' @return Invisibly returns the input object \code{x}.
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' job_id <- lms_download("google/gemma-3-1b")
#' status <- lms_download_status(job_id)
#' print(status)
#' }
print.lms_download_status <- function(x, ...) {
  cli::cli_h3("Download Job: {.val {x$job_id}}")

  # Color-code the status dynamically
  status_col <- switch(
    x$status,
    "downloading" = cli::col_blue,
    "completed" = cli::col_green,
    "already_downloaded" = cli::col_green,
    "failed" = cli::col_red,
    "error" = cli::col_red,
    cli::col_grey
  )

  cli::cli_text("{.strong Status:} ", status_col(x$status))

  # Calculate and format progress
  if (!is.null(x$total_size_bytes) && !is.null(x$downloaded_bytes)) {
    pct <- round((x$downloaded_bytes / x$total_size_bytes) * 100, 1)
    dl_gb <- round(x$downloaded_bytes / (1024^3), 2)
    tot_gb <- round(x$total_size_bytes / (1024^3), 2)

    cli::cli_text("{.strong Progress:} {pct}% ({dl_gb} GB / {tot_gb} GB)")
  }

  # Format speed
  if (!is.null(x$bytes_per_second) && x$bytes_per_second > 0) {
    spd_mb <- round(x$bytes_per_second / (1024^2), 2)
    cli::cli_text("{.strong Speed:} {spd_mb} MB/s")
  }

  # Invisible return so assignment still captures the underlying list
  invisible(x)
}

#' Calculate Expected Scores and Uncertainty from Logprobs
#'
#' Takes a logprobs dataframe (from an \code{lms_chat_result}) and calculates
#' the weighted average score, normalized probabilities, and uncertainty
#' metrics.
#'
#' @param lp_df A dataframe of logprobs (e.g., \code{x$logprobs}).
#' @param scale Numeric vector. The valid labels (e.g., \code{1:5}).
#'
#' @return A named list containing three numeric elements
#'   (\code{expected_value}, \code{weighted_sd}, \code{entropy}) and a
#'   \code{data.frame} named \code{probabilities} with columns \code{label} and
#'   \code{prob}. Returns \code{NULL} if the input dataframe is empty or
#'   invalid.
#'
#' @export
#'
#' @examples
#' # Create a sample logprobs dataframe representing a model's generation step
#' mock_logprobs <- data.frame(
#'   step_token = rep("4", 3),
#'   step_logprob = rep(0, 3),
#'   candidate_token = c("4", "5", "3"),
#'   candidate_logprob = c(-0.105, -2.302, -3.506),
#'   stringsAsFactors = FALSE
#' )
#'
#' # Calculate the expected score and uncertainty metrics
#' lms_score_expected(mock_logprobs, scale = 1:5)
lms_score_expected <- function(lp_df, scale = 1:5) {
  if (is.null(lp_df) || nrow(lp_df) == 0) {
    return(NULL)
  }

  # 1. Isolate the first decision step (where the rating happens)
  # We assume the user followed instructions and the rating is the first token
  first_step <- lp_df$step_token[1]
  candidates <- lp_df[lp_df$step_token == first_step, ]

  # 2. Extract and clean tokens
  # Convert tokens to numbers; non-numeric (like \n) become NA
  nums <- suppressWarnings(as.numeric(candidates$candidate_token))
  valid_idx <- !is.na(nums) & (nums %in% scale)

  if (!any(valid_idx)) {
    cli::cli_abort(
      "No tokens in the top candidates matched the provided scale."
    )
  }

  vals <- nums[valid_idx]
  # Convert logprobs to raw probabilities
  probs <- exp(candidates$candidate_logprob[valid_idx])
  # Normalize so they sum to 1 (re-distributing mass from ignored tokens)
  probs <- probs / sum(probs)

  # 3. Calculations
  # Expected Value: Sum(value * prob)
  ev <- sum(vals * probs)

  # Weighted Variance: Sum(prob * (value - EV)^2)
  var_w <- sum(probs * (vals - ev)^2)
  sd_w <- sqrt(var_w)

  # Shannon Entropy: -Sum(p * log2(p))
  # Measures "surprise" or "confusion" in bits
  entropy <- -sum(probs * log2(probs + 1e-9)) # small epsilon to avoid log(0)

  # 4. Results
  list(
    expected_value = ev,
    weighted_sd = sd_w,
    entropy = entropy,
    probabilities = data.frame(
      label = vals,
      prob = probs,
      stringsAsFactors = FALSE
    )
  )
}

test_that("lms_score_expected calculates correct values for certain outcomes", {
  # Mock: Model is 100% certain the answer is '5'
  # log(1) = 0
  lp_df <- data.frame(
    step_token = c("5", "5"),
    step_logprob = c(0, 0),
    candidate_token = c("5", "4"),
    candidate_logprob = c(0, -20), # -20 is effectively 0 probability
    stringsAsFactors = FALSE
  )

  res <- lms_score_expected(lp_df, scale = 1:5)

  # Expected Value should be 5
  expect_equal(res$expected_value, 5)
  # Weighted SD should be near 0
  expect_lt(res$weighted_sd, 0.01)
  # Entropy should be near 0
  expect_lt(res$entropy, 0.01)
})

test_that("lms_score_expected handles split decisions (bi-modal)", {
  # Mock: Model is split exactly 50/50 between 1 and 5
  # log(0.5) = -0.6931472
  lp_val <- log(0.5)
  lp_df <- data.frame(
    step_token = "1",
    step_logprob = lp_val,
    candidate_token = c("1", "5"),
    candidate_logprob = c(lp_val, lp_val),
    stringsAsFactors = FALSE
  )

  res <- lms_score_expected(lp_df, scale = 1:5)

  # EV should be (1*0.5 + 5*0.5) = 3
  expect_equal(res$expected_value, 3)
  # Entropy should be exactly 1 bit for two equal options
  expect_equal(res$entropy, 1)
  # SD should be 2: sqrt(0.5*(1-3)^2 + 0.5*(5-3)^2) = sqrt(2+2) = 2
  expect_equal(res$weighted_sd, 2)
})

test_that("lms_score_expected filters non-scale tokens", {
  # Mock: Top candidates include a newline or text
  lp_df <- data.frame(
    step_token = "3",
    step_logprob = 0,
    candidate_token = c("3", "\n", "foo"),
    candidate_logprob = c(0, -1, -2),
    stringsAsFactors = FALSE
  )

  # This should pass without error, only using the '3'
  res <- lms_score_expected(lp_df, scale = 1:5)
  expect_equal(res$expected_value, 3)
  expect_equal(nrow(res$probabilities), 1)
})

test_that("lms_score_expected aborts on no valid tokens", {
  lp_df <- data.frame(
    step_token = "A",
    step_logprob = 0,
    candidate_token = c("A", "B"),
    candidate_logprob = c(0, -1),
    stringsAsFactors = FALSE
  )

  expect_error(
    lms_score_expected(lp_df, scale = 1:5),
    "No tokens in the top candidates"
  )
})

#' Check if rlmstudio should be quiet
#'
#' Evaluates whether the user has set the global package option to silence
#' messages, or if a local quiet argument was passed.
#'
#' @param quiet Logical. A local override for the quiet setting.
#'
#' @return Logical.
#'
#' @noRd
#'
#' @examples
#' is_quiet(quiet = TRUE)
#' is_quiet(quiet = FALSE)
is_quiet <- function(quiet = NULL) {
  if (!is.null(quiet)) {
    return(isTRUE(quiet))
  }
  isTRUE(getOption("rlmstudio.quiet", default = FALSE))
}

#' @noRd
#' @examples
#' rlm_alert_info("This is an informational message.")
rlm_alert_info <- function(text, ..., .envir = parent.frame(), quiet = NULL) {
  if (!is_quiet(quiet)) {
    cli::cli_alert_info(text, ..., .envir = .envir)
  }
}

#' @noRd
#' @examples
#' rlm_alert_success("Task completed successfully!")
rlm_alert_success <- function(
  text,
  ...,
  .envir = parent.frame(),
  quiet = NULL
) {
  if (!is_quiet(quiet)) {
    cli::cli_alert_success(text, ..., .envir = .envir)
  }
}

#' @noRd
#' @examples
#' rlm_inform("Standard package information.")
rlm_inform <- function(text, ..., .envir = parent.frame(), quiet = NULL) {
  if (!is_quiet(quiet)) {
    cli::cli_inform(text, ..., .envir = .envir)
  }
}

#' @noRd
#' @examples
#' id <- rlm_progress_step("Initializing process...")
#' rlm_progress_done(id)
rlm_progress_step <- function(
  msg,
  msg_done = msg,
  ...,
  .envir = parent.frame(),
  quiet = NULL
) {
  if (!is_quiet(quiet)) {
    cli::cli_progress_step(msg = msg, msg_done = msg_done, ..., .envir = .envir)
  } else {
    invisible(NULL)
  }
}

#' @noRd
#' @examples
#' id <- rlm_progress_step("Downloading data...")
#' rlm_progress_update(id)
#' rlm_progress_done(id)
rlm_progress_update <- function(
  id = NULL,
  ...,
  .envir = parent.frame(),
  quiet = NULL
) {
  if (!is_quiet(quiet)) {
    cli::cli_progress_update(id = id, ..., .envir = .envir)
  }
}

#' @noRd
#' @examples
#' id <- rlm_progress_step("Finishing setup...")
#' rlm_progress_done(id)
rlm_progress_done <- function(
  id = NULL,
  ...,
  .envir = parent.frame(),
  quiet = NULL
) {
  if (!is_quiet(quiet)) {
    cli::cli_progress_done(id = id, ..., .envir = .envir)
  }
}

#' Build arguments for lms_server_start
#' @noRd
#'
#' @examples
#' build_args_server_start(port = 8080, cors = TRUE)
build_args_server_start <- function(port = NULL, cors = FALSE) {
  args <- c("server", "start")

  if (!is.null(port)) {
    args <- c(args, "--port", as.character(port))
  }

  if (isTRUE(cors)) {
    args <- c(args, "--cors")
  }

  args
}

#' Start the LM Studio local server
#'
#' Launches the LM Studio local server via the CLI, allowing you to interact
#' with loaded models via HTTP API calls.
#'
#' @param port Integer. Port to run the server on. If not provided, LM Studio
#'   uses the last used port.
#' @param cors Logical. Enable CORS support for web application development.
#'   Defaults to FALSE.
#'
#' @seealso [LM Studio CLI Server Start
#'   Documentation](https://lmstudio.ai/docs/cli/serve/server-start)
#'
#' @return Invisibly returns an integer representing the system exit code
#'   (\code{0} for success).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Start server on the default port
#' lms_server_start()
#'
#' # Start server on a custom port with CORS enabled
#' lms_server_start(port = 8080, cors = TRUE)
#' }
lms_server_start <- function(port = NULL, cors = FALSE) {
  args <- build_args_server_start(port = port, cors = cors)

  res <- processx::run(lms_path(), args, error_on_status = FALSE)

  if (res$status == 0) {
    if (!is.null(port)) {
      rlm_alert_success(
        "LM Studio server started successfully on port {.val {port}}."
      )
    } else {
      rlm_alert_success(
        "LM Studio server started successfully on the default port."
      )
    }
  } else {
    cli::cli_abort(
      "Failed to start the LM Studio server. Exit code: {.val {res$status}}."
    )
  }

  invisible(res$status)
}

#' Build arguments for lms_server_stop
#' @noRd
#'
#' @examples
#' build_args_server_stop()
build_args_server_stop <- function() {
  c("server", "stop")
}

#' Stop the LM Studio local server
#'
#' Stops the currently running LM Studio local server via the CLI.
#'
#' @seealso [LM Studio CLI Server Stop
#'   Documentation](https://lmstudio.ai/docs/cli/serve/server-stop)
#'
#' @return Invisibly returns an integer representing the system exit code
#'   (\code{0} for success).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#' lms_server_stop()
#' }
lms_server_stop <- function() {
  args <- build_args_server_stop()

  res <- processx::run(lms_path(), args, error_on_status = FALSE)

  if (res$status == 0) {
    rlm_alert_success("LM Studio server stopped successfully.")
  } else {
    cli::cli_abort(
      "Failed to stop the LM Studio server. Exit code: {.val {res$status}}."
    )
  }

  invisible(res$status)
}

#' Build arguments for lms_server_status
#' @noRd
#'
#' @examples
#' build_args_server_status(json = TRUE, quiet = TRUE)
build_args_server_status <- function(
  json = FALSE,
  verbose = FALSE,
  quiet = FALSE,
  log_level = NULL
) {
  args <- c("server", "status")

  if (isTRUE(json)) {
    args <- c(args, "--json")
  }
  if (isTRUE(verbose)) {
    args <- c(args, "--verbose")
  }
  if (isTRUE(quiet)) {
    args <- c(args, "--quiet")
  }
  if (!is.null(log_level)) {
    args <- c(args, "--log-level", as.character(log_level))
  }

  args
}

#' Check the status of the LM Studio server
#'
#' Displays the current status of the LM Studio local server via the CLI,
#' including whether it is running and its configuration.
#'
#' @param json Logical. Output the status in machine-readable JSON format.
#' @param verbose Logical. Enable detailed logging output.
#' @param quiet Logical. Suppress all logging output.
#' @param log_level Character. The level of logging to use (e.g., "info",
#'   "debug").
#'
#' @details You can only use one logging control flag at a time (`verbose`,
#'   `quiet`, or `log_level`).
#'
#' @seealso [LM Studio CLI Server Status
#'   Documentation](https://lmstudio.ai/docs/cli/serve/server-status)
#'
#' @return By default, returns a character vector containing the raw CLI output.
#'   If \code{json = TRUE} and the \code{jsonlite} package is available, it
#'   returns a parsed list or \code{data.frame} of the status configuration.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' # Get basic status string
#' lms_server_status()
#'
#' # Get status as a parsed JSON data frame
#' lms_server_status(json = TRUE)
#' }
lms_server_status <- function(
  json = FALSE,
  verbose = FALSE,
  quiet = FALSE,
  log_level = NULL
) {
  logging_flags <- sum(c(isTRUE(verbose), isTRUE(quiet), !is.null(log_level)))
  if (logging_flags > 1) {
    cli::cli_warn("Only one logging control flag can be used at a time.")
  }

  args <- build_args_server_status(
    json = json,
    verbose = verbose,
    quiet = quiet,
    log_level = log_level
  )

  res <- processx::run(lms_path(), args, error_on_status = FALSE)
  output <- paste(res$stdout, res$stderr, sep = "\n")

  lines <- strsplit(output, "\r?\n")[[1]]
  lines <- cli::ansi_strip(lines)
  lines <- gsub("\r", "", lines)
  lines <- lines[lines != ""]

  if (isTRUE(json) && requireNamespace("jsonlite", quietly = TRUE)) {
    tryCatch(
      {
        return(jsonlite::fromJSON(paste(lines, collapse = "\n")))
      },
      error = function(e) {
        cli::cli_warn(
          "Failed to parse JSON output. Returning raw character vector instead."
        )
        return(lines)
      }
    )
  }

  return(lines)
}

#' Check if the LM Studio server is reachable
#' @return Logical.
#'
#' @noRd
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' if (is_server_running()) {
#'   message("The LM Studio server is currently active.")
#' }
#' }
is_server_running <- function() {
  tryCatch(
    {
      con <- suppressWarnings(
        socketConnection(host = "localhost", port = 1234, timeout = 0.5)
      )
      close(con)
      TRUE
    },
    error = function(e) FALSE
  )
}

test_that("build_args_server_start constructs correct arguments", {
  expect_equal(build_args_server_start(), c("server", "start"))
  expect_equal(
    build_args_server_start(port = 3000, cors = TRUE),
    c("server", "start", "--port", "3000", "--cors")
  )
})

test_that("build_args_server_status handles logging flags", {
  expect_equal(
    build_args_server_status(json = TRUE, quiet = TRUE),
    c("server", "status", "--json", "--quiet")
  )
})

test_that("lms_server_start handles success and failure", {
  mock_run_success <- function(command, args, error_on_status) list(status = 0)
  local_mocked_bindings(run = mock_run_success, .package = "processx")

  suppressMessages({
    expect_equal(lms_server_start(port = 8080), 0)
  })

  mock_run_fail <- function(command, args, error_on_status) list(status = 1)
  local_mocked_bindings(run = mock_run_fail, .package = "processx")

  expect_error(lms_server_start(), "Failed to start the LM Studio server")
})

test_that("lms_server_status warns on multiple logging flags", {
  mock_run <- function(command, args, error_on_status) {
    list(status = 0, stdout = "ok", stderr = "")
  }
  local_mocked_bindings(run = mock_run, .package = "processx")

  expect_warning(
    lms_server_status(verbose = TRUE, quiet = TRUE),
    "Only one logging control flag can be used at a time"
  )
})

#' Build arguments for lms_daemon_start
#' @noRd
#'
#' @examples
#' build_args_daemon_up()
build_args_daemon_up <- function() {
  c("daemon", "up")
}

#' Start the LM Studio headless daemon
#'
#' Launches the `llmster` daemon in the background via the CLI. This is required
#' in headless environments (such as Linux servers) before loading models or
#' starting the local server.
#'
#' @section Desktop Users: On desktop operating systems (macOS and Windows),
#'   running this command may actually launch the LM Studio desktop application
#'   to act as the backend engine. If the GUI is already open, this function
#'   will simply detect the active instance and return successfully. While safe
#'   to use, desktop users generally do not need to call this function and can
#'   just open the application manually.
#'
#' @seealso [LM Studio Headless Daemon
#'   (llmster)](https://lmstudio.ai/docs/developer/core/headless_llmster)
#'
#' @return Invisibly returns the process object (or 0 if already running).
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_daemon_start()
#' }
lms_daemon_start <- function() {
  args <- build_args_daemon_up()
  res <- processx::run(lms_path(), args, error_on_status = FALSE)

  if (res$status == 0) {
    rlm_alert_success("LM Studio daemon started in the background.")
  } else {
    cli::cli_abort(
      "Failed to start the LM Studio daemon. Exit code: {.val {res$status}}."
    )
  }

  invisible(res$status)
}

#' Check the global status of LM Studio
#'
#' Displays the overall status of the LM Studio backend via the CLI, including loaded
#' models and the server state. This function works regardless of whether
#' the backend was started via the desktop GUI or the headless daemon.
#'
#' @return A character vector of the raw CLI output.
#' @export
#'
#' @examples
#' \dontrun{
#' lms_daemon_status()
#' }
lms_daemon_status <- function() {
  res <- processx::run(lms_path(), "status", error_on_status = FALSE)

  lines <- strsplit(res$stdout, "\r?\n")[[1]]
  lines <- cli::ansi_strip(lines)
  lines <- sub(".*\r", "", lines)
  lines <- lines[lines != ""]

  return(lines)
}

#' Build arguments for lms_daemon_stop
#' @noRd
#'
#' @examples
#' build_args_daemon_down()
build_args_daemon_down <- function() {
  c("daemon", "down")
}

#' Stop the LM Studio headless daemon
#'
#' Stops the `llmster` daemon via the CLI. Use this to clean up system resources when
#' you are completely finished using LM Studio in headless mode.
#'
#' @section Desktop Users:
#' If the daemon is currently being managed by the LM Studio desktop
#' application, this function will fail. The CLI intentionally prevents
#' programmatic shutdowns of the GUI to avoid disrupting visual sessions.
#' In this scenario, you must close the desktop application manually.
#'
#' @param force Logical. If `TRUE`, attempts to stop the local server before
#'   shutting down the daemon. The daemon cannot be stopped while the server
#'   is actively running. Defaults to `FALSE`.
#'
#' @return Invisibly returns the system exit code (0 for success).
#' @export
#'
#' @examples
#' \dontrun{
#' lms_daemon_stop(force = TRUE)
#' }
lms_daemon_stop <- function(force = FALSE) {
  if (isTRUE(force)) {
    tryCatch(lms_server_stop(), error = function(e) NULL)
  }

  args <- build_args_daemon_down()
  res <- processx::run(lms_path(), args, error_on_status = FALSE)

  if (res$status == 0) {
    rlm_alert_success("LM Studio daemon stopped successfully.")
    return(invisible(TRUE))
  } else {
    err_msg <- trimws(res$stderr)
    if (err_msg == "") {
      err_msg <- trimws(res$stdout)
    }
    if (err_msg == "") {
      err_msg <- "Unknown CLI error."
    }

    # Catch the GUI conflict and exit gracefully
    if (grepl("part of LM Studio", err_msg, ignore.case = TRUE)) {
      rlm_alert_info(
        "The daemon is managed by the LM Studio GUI and will remain running."
      )
      return(invisible(FALSE))
    }

    # Catch the "already stopped" scenario and exit gracefully
    if (grepl("not running", err_msg, ignore.case = TRUE)) {
      rlm_alert_info("The LM Studio daemon is already stopped.")
      return(invisible(TRUE))
    }

    # For all other errors, abort as usual
    cli::cli_abort(c(
      "Failed to stop the LM Studio daemon. Exit code: {.val {res$status}}.",
      "x" = "CLI output: {.val {err_msg}}",
      "i" = "Hint: If the server is still running, try `lms_daemon_stop(force = TRUE)` or run `lms_server_stop()` first."
    ))
  }
}

#' Run code with the LM Studio daemon active
#'
#' Temporarily starts the LM Studio headless daemon, executes the provided
#' R expression, and then gracefully shuts the daemon and any active servers
#' down. This is ideal for automated scripts and pipelines.
#'
#' @section Desktop Users:
#' Be cautious using this wrapper if you already have the LM Studio GUI open.
#' While the setup phase (`lms_daemon_start`) will succeed, the teardown phase
#' (`lms_daemon_stop`) will fail because the CLI prevents programmatic shutdowns
#' of the graphical interface. This wrapper is best reserved for strictly
#' headless environments or fully automated scripts.
#'
#' @param code An R expression to execute while the daemon is running.
#'
#' @return The result of the evaluated code.
#' @export
#'
#' @examples
#' \dontrun{
#' result <- with_lms_daemon({
#'   lms_load("llama-3.1-8b")
#'   lms_chat("llama-3.1-8b", input = "Hello world!")
#' })
#' }
with_lms_daemon <- function(code) {
  lms_daemon_start()

  on.exit(lms_daemon_stop(force = TRUE), add = TRUE)

  force(code)
}

test_that("lms_daemon_start handles successful execution", {
  mock_run <- function(command, args, error_on_status) {
    list(status = 0, stdout = "Success", stderr = "")
  }

  testthat::local_mocked_bindings(run = mock_run, .package = "processx")

  # Suppress the cli alert for cleaner test output
  suppressMessages({
    res <- lms_daemon_start()
  })

  expect_equal(res, 0)
})

test_that("lms_daemon_start handles errors", {
  mock_run_fail <- function(command, args, error_on_status) {
    list(status = 1, stdout = "", stderr = "Error starting daemon")
  }
  local_mocked_bindings(run = mock_run_fail, .package = "processx")

  expect_error(lms_daemon_start(), "Failed to start the LM Studio daemon")
})

test_that("lms_daemon_status parses multi-line output correctly", {
  mock_run_status <- function(command, args, error_on_status) {
    list(status = 0, stdout = "Status OK\n\rLoaded models: 2\n", stderr = "")
  }
  local_mocked_bindings(run = mock_run_status, .package = "processx")

  res <- lms_daemon_status()
  expect_equal(res, c("Status OK", "Loaded models: 2"))
})

test_that("lms_daemon_stop exits gracefully when managed by GUI", {
  mock_run_gui <- function(command, args, error_on_status) {
    list(status = 1, stdout = "", stderr = "part of LM Studio")
  }
  local_mocked_bindings(run = mock_run_gui, .package = "processx")

  # Expect the informational message and a return value of FALSE
  expect_message(
    res <- lms_daemon_stop(),
    "managed by the LM Studio GUI"
  )
  expect_false(res)
})

test_that("lms_daemon_stop handles force argument and generic failures", {
  mock_run_fail <- function(command, args, error_on_status) {
    list(status = 1, stdout = "", stderr = "Some unknown system error")
  }
  local_mocked_bindings(run = mock_run_fail, .package = "processx")

  expect_error(
    lms_daemon_stop(),
    "Failed to stop the LM Studio daemon"
  )
})

#' Check if LM Studio CLI is installed
#'
#' @return A logical scalar: \code{TRUE} if the \code{lms} executable is found
#'   on the system path, and \code{FALSE} otherwise.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' has_lms()
#' }
has_lms <- function() {
  Sys.which("lms") != ""
}

#' Check if the installed LM Studio CLI meets the minimum requirement
#'
#' @param min_version Character string of the required version. Default is
#'   "0.4.0".
#'
#' @return A logical scalar: \code{TRUE} if the LM Studio CLI version meets or
#'   exceeds the specified \code{min_version}, and \code{FALSE} otherwise.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' check_lms_version("0.4.0")
#' }
check_lms_version <- function(min_version = "0.4.0") {
  if (!has_lms()) {
    cli::cli_alert_danger("LM Studio CLI is not installed.")
    return(FALSE)
  }

  tryCatch(
    {
      result <- processx::run(
        lms_path(),
        args = "--version",
        error_on_status = FALSE
      )
      output <- paste(result$stdout, result$stderr)

      if (grepl("CLI commit:", output, ignore.case = TRUE)) {
        rlm_alert_success(
          "LM Studio CLI is using the modern architecture (0.4.0+)."
        )
        return(TRUE)
      }

      version_string <- regmatches(
        output,
        regexpr("[0-9]+\\.[0-9]+\\.[0-9]+", output)
      )

      if (length(version_string) == 0) {
        cli::cli_alert_warning(c(
          "Could not parse the LM Studio CLI version. Output was: ",
          "{.val {trimws(output)}}"
        ))
        return(FALSE)
      }

      if (numeric_version(version_string) >= numeric_version(min_version)) {
        rlm_alert_success(
          "LM Studio CLI version {.val {version_string}} meets the requirement ({.val {min_version}})."
        )
        return(TRUE)
      } else {
        cli::cli_alert_danger(
          "LM Studio CLI version {.val {version_string}} is too old. Minimum required is {.val {min_version}}."
        )
        return(FALSE)
      }
    },
    error = function(e) {
      cli::cli_abort(c(
        "x" = "Failed to check LM Studio CLI version.",
        "i" = "Error message: {.val {e$message}}"
      ))
    }
  )
}

#' Help the user install or update LM Studio
#'
#' This function provides two methods for setting up LM Studio on your system.
#' The "browser" method opens the official download page for the LM Studio
#' desktop application (GUI). The "headless" method runs an automated
#' installation script to install the \code{llmster} daemon and CLI, which is
#' suitable for servers, containers, or users who prefer a GUI-less environment.
#'
#' @param method Character. Either "browser" (opens the GUI download page) or
#'   "headless" (installs the \code{llmster} daemon via script).
#'
#' @return Invisibly returns \code{TRUE} upon successful completion. This
#'   function is primarily utilized for its side effects of opening a web
#'   browser or executing system installation commands.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Open your default web browser to the download page
#' install_lmstudio(method = "browser")
#'
#' # Attempt automatic headless installation via the command line
#' install_lmstudio(method = "headless")
#' }
install_lmstudio <- function(method = c("browser", "headless")) {
  method <- match.arg(method)

  # Check for existence AND version before proceeding
  if (has_lms() && check_lms_version("0.4.0")) {
    rlm_alert_success("Your LM Studio setup is ready to go!")
    return(invisible(TRUE))
  }

  if (method == "browser") {
    rlm_alert_info(
      "Opening the LM Studio download page in your default browser..."
    )
    utils::browseURL("https://lmstudio.ai/download")
    cli::cli_alert_warning(
      "Please install or update the software, restart R, and try again."
    )
  } else if (method == "headless") {
    # CRAN Compliance: Require interactive consent or explicit environment variable
    if (
      !interactive() &&
        !isTRUE(as.logical(Sys.getenv("RLMSTUDIO_ALLOW_INSTALL", "FALSE")))
    ) {
      cli::cli_abort(c(
        "Installation requires an interactive session to grant permission.",
        "i" = "To install automatically in non-interactive scripts or CI/CD, set the {.envvar RLMSTUDIO_ALLOW_INSTALL} environment variable to {.val TRUE}."
      ))
    }

    if (interactive()) {
      consent <- utils::askYesNo(
        "This will download and install the LM Studio CLI to your system. Do you want to proceed?"
      )
      if (!isTRUE(consent)) {
        cli::cli_abort("Installation cancelled by user.")
      }
    }

    os <- Sys.info()[["sysname"]]
    rlm_progress_step("Downloading and installing LM Studio CLI...")

    tryCatch(
      {
        if (os %in% c("Darwin", "Linux")) {
          if (Sys.which("curl") == "") {
            cli::cli_abort(
              "The system command {.val curl} is required but was not found."
            )
          }

          res <- processx::run(
            command = "bash",
            args = c(
              "-c",
              "set -o pipefail; curl -fsSL https://lmstudio.ai/install.sh | bash"
            ),
            echo = FALSE,
            stderr_to_stdout = TRUE,
            error_on_status = FALSE
          )
        } else if (os == "Windows") {
          res <- processx::run(
            command = "powershell",
            args = c("-Command", "irm https://lmstudio.ai/install.ps1 | iex"),
            echo = FALSE,
            stderr_to_stdout = TRUE,
            error_on_status = FALSE
          )
        } else {
          cli::cli_abort(
            "Automatic installation is not supported for this operating system: {.val {os}}."
          )
        }

        if (res$status == 0) {
          rlm_progress_done()
          rlm_alert_success("LM Studio CLI installed successfully.")
        } else {
          cli::cli_progress_cleanup()
          cli::cli_abort(c(
            "x" = "Headless installation failed. Exit code: {.val {res$status}}.",
            "i" = "CLI output: {.val {trimws(res$stdout)}}"
          ))
        }

        rlm_alert_info(
          "In a headless environment, remember to start the daemon using {.fn lms_daemon_start} and the server using {.fn lms_server_start} before loading models."
        )
      },
      error = function(e) {
        cli::cli_progress_cleanup()
        cli::cli_abort(c(
          "x" = "Headless installation failed.",
          "i" = "Error message: {.val {e$message}}"
        ))
      }
    )
  }

  return(invisible(TRUE))
}

#' Load a model via REST API
#'
#' @param model Character. Unique identifier for the model to load.
#' @param context_length Integer. Maximum number of tokens that the model will
#'   consider.
#' @param eval_batch_size Integer. Number of input tokens to process together in
#'   a single batch during evaluation.
#' @param flash_attention Logical. Whether to optimize attention computation.
#' @param num_experts Integer. Number of experts to use during inference for MoE
#'   models.
#' @param offload_kv_cache_to_gpu Logical. Whether KV cache is offloaded to GPU
#'   memory.
#' @param echo_load_config Logical. If \code{TRUE}, echoes the final load
#'   configuration in the response.
#' @param force Logical. If \code{TRUE}, bypasses the check for currently loaded
#'   models and requests a new instance from the server. Note that this does not
#'   overwrite or replace the existing model; it loads a second concurrent
#'   instance into VRAM. Defaults to \code{FALSE}.
#' @param host Character. The host address of the local server. Defaults to
#'   "http://localhost:1234".
#' @param ... Additional arguments passed to the API request body (useful for
#'   future API parameters).
#'
#' @seealso [LM Studio Load Model
#'   API](https://lmstudio.ai/docs/developer/rest/load)
#'
#' @return Invisibly returns a character string of the loaded model identifier
#'   upon success. If \code{echo_load_config = TRUE}, it instead invisibly
#'   returns a list containing the model's detailed load configuration.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#' lms_download("google/gemma-3-1b")
#'
#' # Load a model with default settings
#' lms_load("google/gemma-3-1b")
#'
#' # Load a model with custom context length and flash attention enabled
#' lms_load("google/gemma-3-1b", context_length = 8192, flash_attention = TRUE)
#' }
lms_load <- function(
  model,
  context_length = NULL,
  eval_batch_size = NULL,
  flash_attention = NULL,
  num_experts = NULL,
  offload_kv_cache_to_gpu = NULL,
  echo_load_config = FALSE,
  force = FALSE,
  host = "http://localhost:1234",
  ...
) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  # Check if the model is already loaded to prevent redundant API calls
  if (!isTRUE(force)) {
    active_models <- list_models(
      loaded = TRUE,
      detailed = TRUE,
      quiet = TRUE,
      host = host
    )

    if (nrow(active_models) > 0 && model %in% active_models$key) {
      rlm_alert_info(
        "Model {.val {model}} is already loaded. Use {.code force = TRUE} to load an additional instance."
      )
      if (isTRUE(echo_load_config)) {
        cli::cli_alert_warning(
          "Cannot echo load config because the model was already loaded."
        )
      }
      return(invisible(model))
    }
  }

  # 1. Build the explicit body based on current known parameters
  body <- list(
    model = model,
    context_length = if (!is.null(context_length)) {
      as.integer(context_length)
    } else {
      NULL
    },
    eval_batch_size = if (!is.null(eval_batch_size)) {
      as.integer(eval_batch_size)
    } else {
      NULL
    },
    flash_attention = if (!is.null(flash_attention)) {
      as.logical(flash_attention)
    } else {
      NULL
    },
    num_experts = if (!is.null(num_experts)) as.integer(num_experts) else NULL,
    offload_kv_cache_to_gpu = if (!is.null(offload_kv_cache_to_gpu)) {
      as.logical(offload_kv_cache_to_gpu)
    } else {
      NULL
    },
    echo_load_config = if (isTRUE(echo_load_config)) TRUE else NULL
  )

  # 2. Merge dots into the body (allows for future/undocumented API parameters)
  body <- utils::modifyList(Filter(Negate(is.null), body), list(...))

  rlm_progress_step(
    msg = "Loading model: {.val {model}}...",
    msg_done = "Model {.val {model}} loaded and verified."
  )

  resp <- lms_client(host) |>
    httr2::req_url_path("api/v1/models/load") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    resp_data <- httr2::resp_body_json(resp)
    if (identical(resp_data$status, "loaded")) {
      if (isTRUE(echo_load_config)) {
        return(invisible(resp_data$load_config))
      }
      return(invisible(model))
    }
  }

  err_msg <- tryCatch(
    {
      err_json <- httr2::resp_body_json(resp)
      if (!is.null(err_json$error$message)) {
        err_json$error$message
      } else if (!is.null(err_json$error)) {
        err_json$error
      } else {
        httr2::resp_body_string(resp)
      }
    },
    error = function(e) httr2::resp_body_string(resp)
  )

  if (err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }
  cli::cli_abort(c("x" = "API Load Failed: {err_msg}"), call = NULL)
}

test_that("lms_load builds body with correct integer/logical conversions", {
  local_mocked_bindings(is_server_running = function() TRUE)

  # We capture the request object to inspect the built JSON body
  captured_req <- NULL
  mock_perform <- function(req) {
    captured_req <<- req
    # Return a fake successful response
    httr2::response(
      status_code = 200,
      body = charToRaw('{"status": "loaded"}'),
      headers = list(`Content-Type` = "application/json")
    )
  }
  local_mocked_bindings(req_perform = mock_perform, .package = "httr2")

  suppressMessages({
    lms_load(
      model = "test-model",
      context_length = "2048", # string that should become integer
      flash_attention = 1, # numeric that should become logical
      custom_param = "extra" # testing the dots (...) modification
    )
  })

  body_data <- captured_req$body$data
  expect_equal(body_data$model, "test-model")
  expect_equal(body_data$context_length, 2048L)
  expect_true(body_data$flash_attention)
  expect_equal(body_data$custom_param, "extra")
})

#' Get the absolute path to the LMS executable
#'
#' Locates the LM Studio CLI (`lms`) on your system. It checks the
#' `RLMSTUDIO_LMS_PATH` environment variable first, then the system `PATH`, and
#' finally common installation directories.
#'
#' @return A character string specifying the absolute file path to the LM Studio
#'   executable (\code{lms}) on the user's system.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_path()
#' }
lms_path <- function() {
  # 1. Allow an explicit override via environment variable
  custom_path <- Sys.getenv("RLMSTUDIO_LMS_PATH")
  if (custom_path != "" && file.exists(custom_path)) {
    return(custom_path)
  }

  # 2. Check the standard system PATH
  cmd <- Sys.which("lms")
  if (cmd != "") {
    return(unname(cmd))
  }

  # 3. Fallback to checking common installation directories
  os <- Sys.info()[["sysname"]]
  home <- path.expand("~")

  common_paths <- if (os == "Windows") {
    c(file.path(Sys.getenv("LOCALAPPDATA"), "LM-Studio", "bin", "lms.exe"))
  } else {
    c(
      file.path(home, ".lmstudio", "bin", "lms"),
      file.path(home, ".cache", "lm-studio", "bin", "lms"),
      file.path(home, ".local", "bin", "lms")
    )
  }

  for (p in common_paths) {
    if (file.exists(p)) {
      return(p)
    }
  }

  # 4. Abort if all checks fail
  cli::cli_abort(c(
    "x" = "The LM Studio CLI ({.val lms}) was not found on your system.",
    "i" = "Please install it or run {.fun install_lmstudio} to configure your environment."
  ))
}

test_that("lms_path respects the RLMSTUDIO_LMS_PATH env var", {
  withr::with_envvar(c(RLMSTUDIO_LMS_PATH = "/custom/path/lms"), {
    local_mocked_bindings(file.exists = function(x) TRUE, .package = "base")
    expect_equal(lms_path(), "/custom/path/lms")
  })
})

test_that("lms_path aborts when CLI is completely missing", {
  withr::with_envvar(c(RLMSTUDIO_LMS_PATH = ""), {
    local_mocked_bindings(Sys.which = function(x) "", .package = "base")
    local_mocked_bindings(file.exists = function(x) FALSE, .package = "base")

    expect_error(lms_path(), "was not found on your system")
  })
})

skip_if_no_lms <- function() {
  if (!has_lms()) {
    testthat::skip("LM Studio CLI is not installed.")
  }
}

skip_if_no_server <- function() {
  if (!is_server_running()) {
    testthat::skip("LM Studio local server is not running.")
  }
}

test_that("End-to-end model load and chat works", {
  testthat::skip_on_cran()

  local_mocked_bindings(
    has_lms = function() TRUE,
    is_server_running = function() TRUE
  )

  skip_if_no_lms()
  skip_if_no_server()

  test_model <- "google/gemma-3-1b"

  # Safe teardown fallback (silences the error if it runs unmocked)
  on.exit(
    {
      try(lms_unload(test_model), silent = TRUE)
    },
    add = TRUE
  )

  httptest2::with_mock_dir("integration_e2e", {
    load_res <- lms_load(test_model)

    models <- list_models(loaded = TRUE, quiet = TRUE)

    chat_res <- lms_chat(
      model = test_model,
      input = "Reply with exactly the word: 'Received'."
    )

    # Explicitly unload inside the mock so httptest2 records it
    try(lms_unload(test_model), silent = TRUE)
  })

  expect_equal(load_res, test_model)
  expect_true(test_model %in% models$key)
  expect_type(chat_res, "character")
  expect_true(nchar(chat_res) > 0)
})

test_that("Full Integration: Download, Load, and Rate", {
  local_mocked_bindings(
    has_lms = function() TRUE,
    is_server_running = function() TRUE
  )

  skip_if_no_lms()
  skip_if_no_server()

  target_model <- "google/gemma-3-1b"

  # Safe teardown fallback (silences the error if it runs unmocked)
  on.exit(
    {
      try(lms_unload(target_model), silent = TRUE)
    },
    add = TRUE
  )

  httptest2::with_mock_dir("chat_integration", {
    lms_download(target_model)
    lms_load(target_model, flash_attention = TRUE)

    reviews <- c(
      "The food was absolutely incredible!",
      "Terrible service, I will never go back."
    )
    sys_prompt <- "Rate on a scale of 1 to 5. Respond with ONLY the integer."

    results_df <- lms_chat_batch(
      model = target_model,
      inputs = reviews,
      system_prompt = sys_prompt,
      format = "data.frame",
      api_type = "openresponses",
      logprobs = TRUE,
      top_logprobs = 5,
      temperature = 1.0,
      quiet = TRUE
    )

    # Explicitly unload inside the mock so httptest2 records it
    try(lms_unload(target_model), silent = TRUE)
  })

  expect_s3_class(results_df, "data.frame")
  expect_true("logprobs" %in% names(results_df))
  expect_s3_class(results_df$logprobs[[1]], "data.frame")

  results_df$expected_rating <- sapply(results_df$logprobs, function(lp_df) {
    if (is.null(lp_df)) {
      return(NA_real_)
    }

    candidates <- lp_df[lp_df$step_token == lp_df$step_token[1], ]

    nums <- suppressWarnings(as.numeric(candidates$candidate_token))
    valid <- !is.na(nums) & nums %in% 1:5

    if (!any(valid)) {
      return(NA_real_)
    }

    vals <- nums[valid]
    probs <- exp(candidates$candidate_logprob[valid])
    probs <- probs / sum(probs)

    sum(vals * probs)
  })

  expect_true(all(
    results_df$expected_rating >= 1 & results_df$expected_rating <= 5
  ))
})

test_that("lms_chat routes correctly to openresponses and creates S3 class", {
  local_mocked_bindings(is_server_running = function() TRUE)

  fake_body <- list(
    output = list(list(
      content = list(list(
        text = "5",
        logprobs = list(
          list(
            token = "5",
            logprob = 0,
            top_logprobs = list(list(token = "5", logprob = 0))
          )
        )
      ))
    ))
  )

  local_mocked_bindings(
    req_perform = function(req, ...) {
      httr2::response(
        status_code = 200,
        headers = list("Content-Type" = "application/json"),
        body = charToRaw(jsonlite::toJSON(fake_body, auto_unbox = TRUE))
      )
    },
    .package = "httr2"
  )

  res <- lms_chat(
    model = "test-model",
    input = "test input",
    api_type = "openresponses",
    logprobs = TRUE
  )

  expect_s3_class(res, "lms_chat_result")
  expect_equal(res$text, "5")
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @section Options:
#' \itemize{
#'   \item \code{rlmstudio.quiet}: A logical value. If \code{TRUE}, suppresses informational console messages and progress bars across the package. Defaults to \code{FALSE}.
#' }
## usethis namespace: end
NULL

#' List available models
#'
#' Retrieves a list of models available on your system via the LM Studio REST
#' API.
#'
#' @param loaded Logical. If \code{TRUE}, returns only currently loaded models.
#'   Defaults to \code{FALSE}.
#' @param type Character vector. The types of models to include. Defaults to
#'   \code{c("llm", "embedding")}.
#' @param detailed Logical. Show all information about each model. Defaults to
#'   \code{FALSE}.
#' @param quiet Logical. If \code{TRUE}, suppresses informative console
#'   messages. Defaults to \code{FALSE}.
#' @param host Character. The host address of the local server.
#'
#' @seealso [LM Studio List Models
#'   API](https://lmstudio.ai/docs/developer/rest/list)
#'
#' @return A \code{data.frame} containing information about the available
#' models. By default, it includes columns for \code{state}, \code{type},
#' \code{display_name}, \code{key}, \code{architecture}, and \code{size_gb}.
#' If \code{detailed = TRUE}, it returns a comprehensive \code{data.frame}
#' including all raw metadata columns provided by the API. Returns an empty
#' \code{data.frame} if no models match the criteria.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#' lms_download("google/gemma-3-1b")
#' lms_load("google/gemma-3-1b")
#'
#' # List all downloaded models
#' list_models()
#'
#' # List only currently loaded models
#' list_models(loaded = TRUE)
#'
#' # Get detailed information about loaded text models
#' list_models(loaded = TRUE, type = "llm", detailed = TRUE)
#' }
list_models <- function(
  loaded = FALSE,
  type = c("llm", "embedding"),
  detailed = FALSE,
  quiet = FALSE,
  host = "http://localhost:1234"
) {
  if (!is_server_running()) {
    if (!quiet) {
      cli::cli_alert_danger(
        "The LM Studio server is not running. Run {.fn lms_server_start} first."
      )
    }
    return(invisible(data.frame()))
  }

  resp <- lms_client(host) |>
    httr2::req_url_path("api/v1/models") |>
    httr2::req_perform()

  raw_content <- httr2::resp_body_string(resp)
  full_data <- jsonlite::fromJSON(raw_content, simplifyDataFrame = TRUE)
  df <- full_data$models

  if (is.null(df) || nrow(df) == 0) {
    if (!quiet) {
      rlm_inform(c("i" = "No models found on host {.url {host}}."))
    }
    return(invisible(data.frame()))
  }

  # Apply type filters
  df <- df[df$type %in% type, , drop = FALSE]

  # Evaluate load state for all matching models
  if (nrow(df) > 0) {
    df$state <- ifelse(
      vapply(
        df$loaded_instances,
        function(x) {
          if (is.data.frame(x)) nrow(x) > 0 else length(x) > 0
        },
        logical(1)
      ),
      "loaded",
      "unloaded"
    )

    # Apply loaded filter
    if (isTRUE(loaded)) {
      df <- df[df$state == "loaded", , drop = FALSE]
    }
  }

  if (nrow(df) == 0) {
    if (!quiet) {
      rlm_inform(c(
        "!" = "No models found matching criteria: loaded = {.val {loaded}}, type = {.val {type}}."
      ))
    }
    return(invisible(data.frame()))
  }

  # Format size for readability
  if ("size_bytes" %in% names(df)) {
    df$size_gb <- round(df$size_bytes / (1024^3), 2)
  }

  # Clean up and select columns
  if (!isTRUE(detailed)) {
    core_cols <- c(
      "state",
      "type",
      "display_name",
      "key",
      "architecture",
      "size_gb"
    )
    available_cols <- intersect(core_cols, names(df))
    df <- df[, available_cols, drop = FALSE]
  }

  return(df)
}

test_that("list_models returns a formatted data frame", {
  # Mock the server check to always return TRUE for this test
  testthat::local_mocked_bindings(is_server_running = function() TRUE)

  httptest2::with_mock_dir("list_models", {
    # The first time this runs, you must have the server on.
    # It will save the HTTP response to tests/testthat/list_models/
    res <- list_models(host = "http://localhost:1234")

    expect_s3_class(res, "data.frame")
    expect_true(all(c("state", "type", "display_name", "key") %in% names(res)))
  })
})

#' Chat Completion with LM Studio
#'
#' Send a prompt to a locally running LM Studio model. This wrapper
#' automatically routes your request to the appropriate subfunction based on the
#' selected API type.
#'
#' @param model Character. The name of the loaded model.
#' @param input Character. The user prompt to send to the model.
#' @param system_prompt Character. An optional system prompt to guide model
#'   behavior.
#' @param host Character. The base URL of the LM Studio server. Default is
#'   "http://localhost:1234".
#' @param api_type Character. The LM Studio API endpoint to use. Options are
#'   "openresponses" (default), "openai", or "native".
#' @param logprobs Logical. Whether to return the log probabilities of the
#'   generated tokens. Default is FALSE.
#' @param simplify Logical. If TRUE, extracts the core text response. Default is
#'   TRUE.
#' @param ... Additional arguments passed to the selected API body.
#' @return Depending on the arguments provided:
#' \itemize{
#'   \item If \code{simplify = FALSE}, returns a parsed list of the raw JSON response.
#'   \item If \code{simplify = TRUE} and \code{logprobs = FALSE}, returns a single character string containing the model's text response.
#'   \item If \code{simplify = TRUE} and \code{logprobs = TRUE} (and the chosen API type supports it), returns an object of class \code{lms_chat_result} containing both the text and a data.frame of token probabilities.
#' }
#' @export
lms_chat <- function(
  model,
  input,
  system_prompt = NULL,
  host = "http://localhost:1234",
  api_type = c("openresponses", "openai", "native"),
  logprobs = FALSE,
  simplify = TRUE,
  ...
) {
  api_type <- match.arg(api_type)

  if (api_type == "openresponses") {
    return(lms_chat_openresponses(
      model = model,
      input = input,
      instructions = system_prompt,
      host = host,
      logprobs = logprobs,
      simplify = simplify,
      ...
    ))
  }

  if (api_type == "openai") {
    msgs <- list()
    if (!is.null(system_prompt)) {
      msgs[[length(msgs) + 1]] <- list(role = "system", content = system_prompt)
    }
    msgs[[length(msgs) + 1]] <- list(role = "user", content = input)

    return(lms_chat_openai(
      model = model,
      messages = msgs,
      host = host,
      logprobs = logprobs,
      simplify = simplify,
      ...
    ))
  }

  if (api_type == "native") {
    if (isTRUE(logprobs)) {
      cli::cli_warn(
        "The 'native' API type does not support logprobs. Ignoring argument."
      )
    }
    return(lms_chat_native(
      model = model,
      input = input,
      system_prompt = system_prompt,
      host = host,
      simplify = simplify,
      ...
    ))
  }
}

#' Chat Completion via OpenResponses API
#'
#' Direct interface to LM Studio's OpenResponses endpoint. Supports logprobs and
#' custom instructions.
#'
#' @param model Character. The loaded model name.
#' @param input Character. The user prompt.
#' @param instructions Character. Optional system instructions.
#' @param host Character. Server URL.
#' @param logprobs Logical. Whether to return token probabilities.
#' @param simplify Logical. If TRUE, parses output to text and dataframe. If
#'   FALSE, returns raw list.
#' @param ... Additional API arguments (e.g., top_logprobs, temperature).
#' @return If \code{simplify = FALSE}, returns a list representing the raw JSON
#'   response. Otherwise, returns a character string containing the generated
#'   text. If \code{logprobs = TRUE}, returns an object of class
#'   \code{lms_chat_result} incorporating both the text and probability data.
#' @export
lms_chat_openresponses <- function(
  model,
  input,
  instructions = NULL,
  host = "http://localhost:1234",
  logprobs = FALSE,
  simplify = TRUE,
  ...
) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  body <- list(model = model, input = input, instructions = instructions)
  if (isTRUE(logprobs)) {
    body$include <- list("message.output_text.logprobs")
  }

  body <- Filter(Negate(is.null), body)
  body <- utils::modifyList(body, list(...))

  resp <- lms_client(host) |>
    httr2::req_url_path("v1/responses") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    resp_data <- httr2::resp_body_json(resp)
    if (!isTRUE(simplify)) {
      return(resp_data)
    }

    content <- resp_data$output[[1]]$content[[1]]

    if (
      isTRUE(logprobs) &&
        !is.null(content$logprobs) &&
        length(content$logprobs) > 0
    ) {
      logprobs_df <- do.call(
        rbind,
        lapply(content$logprobs, function(step) {
          step_tok <- if (is.null(step$token)) NA_character_ else step$token
          step_lp <- if (is.null(step$logprob)) NA_real_ else step$logprob

          if (is.null(step$top_logprobs) || length(step$top_logprobs) == 0) {
            return(data.frame(
              step_token = step_tok,
              step_logprob = step_lp,
              candidate_token = NA_character_,
              candidate_logprob = NA_real_,
              stringsAsFactors = FALSE
            ))
          }

          do.call(
            rbind,
            lapply(step$top_logprobs, function(cand) {
              data.frame(
                step_token = step_tok,
                step_logprob = step_lp,
                candidate_token = if (is.null(cand$token)) {
                  NA_character_
                } else {
                  cand$token
                },
                candidate_logprob = if (is.null(cand$logprob)) {
                  NA_real_
                } else {
                  cand$logprob
                },
                stringsAsFactors = FALSE
              )
            })
          )
        })
      )
      rownames(logprobs_df) <- NULL

      # Use S3 Constructor and Validator
      return(validate_lms_chat_result(
        new_lms_chat_result(text = content$text, logprobs = logprobs_df)
      ))
    }

    return(content$text)
  }

  err_msg <- tryCatch(
    {
      httr2::resp_body_json(resp)$error$message
    },
    error = function(e) ""
  )
  if (is.null(err_msg) || err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }
  cli::cli_abort(c("x" = "OpenResponses Failed: {err_msg}"), call = NULL)
}

#' Chat Completion via OpenAI Compatibility API
#'
#' Direct interface to LM Studio's OpenAI-compatible endpoint. Uses the messages
#' array format.
#'
#' @param model Character. The loaded model name.
#' @param messages List. A structured list of role and content pairs.
#' @param host Character. Server URL.
#' @param logprobs Logical. Whether to request logprobs (currently stubbed by LM
#'   Studio).
#' @param simplify Logical. If TRUE, parses output to text.
#' @param ... Additional API arguments.
#' @return If \code{simplify = FALSE}, returns a list representing the raw JSON
#'   response. Otherwise, returns a character string containing the generated
#'   text. If \code{logprobs = TRUE}, it returns an \code{lms_chat_result}
#'   object with the log probabilities populated as \code{NULL} since they are
#'   currently stubbed in the LM Studio OpenAI endpoint.
#' @export
lms_chat_openai <- function(
  model,
  messages,
  host = "http://localhost:1234",
  logprobs = FALSE,
  simplify = TRUE,
  ...
) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  body <- list(model = model, messages = messages)
  if (isTRUE(logprobs)) {
    body$logprobs <- TRUE
  }

  body <- Filter(Negate(is.null), body)
  body <- utils::modifyList(body, list(...))

  resp <- lms_client(host) |>
    httr2::req_url_path("v1/chat/completions") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    resp_data <- httr2::resp_body_json(resp)
    if (!isTRUE(simplify)) {
      return(resp_data)
    }

    res_text <- resp_data$choices[[1]]$message$content

    if (isTRUE(logprobs)) {
      # Return S3 object with NULL logprobs (since OpenAI endpoint is a stub in LM Studio)
      return(validate_lms_chat_result(
        new_lms_chat_result(text = res_text, logprobs = NULL)
      ))
    }
    return(res_text)
  }

  err_msg <- tryCatch(
    {
      httr2::resp_body_json(resp)$error$message
    },
    error = function(e) ""
  )
  if (is.null(err_msg) || err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }
  cli::cli_abort(c("x" = "OpenAI API Failed: {err_msg}"), call = NULL)
}

#' Chat Completion via Native API
#'
#' Direct interface to LM Studio's v1 Native endpoint. Optimized for stateful chats and hardware control.
#'
#' @param model Character. The loaded model name.
#' @param input Character. The user prompt.
#' @param system_prompt Character. Optional system prompt.
#' @param host Character. Server URL.
#' @param simplify Logical. If TRUE, parses output to text.
#' @param ... Additional API arguments.
#' @return If \code{simplify = FALSE}, returns a list representing the raw JSON
#'   response. If \code{simplify = TRUE}, returns a character string containing
#'   the model's text output.
#' @export
lms_chat_native <- function(
  model,
  input,
  system_prompt = NULL,
  host = "http://localhost:1234",
  simplify = TRUE,
  ...
) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  body <- list(model = model, input = input, system_prompt = system_prompt)
  body <- Filter(Negate(is.null), body)

  # Check if user tried to pass logprobs in dots and warn them
  dots <- list(...)
  if (isTRUE(dots$logprobs)) {
    cli::cli_warn(
      "The native API does not support logprobs. Ignoring argument."
    )
    dots$logprobs <- NULL
  }
  body <- utils::modifyList(body, dots)

  resp <- lms_client(host) |>
    httr2::req_url_path("api/v1/chat") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    resp_data <- httr2::resp_body_json(resp)
    if (!isTRUE(simplify)) {
      return(resp_data)
    }
    return(resp_data$output[[1]]$content)
  }

  err_msg <- tryCatch(
    {
      httr2::resp_body_json(resp)$error
    },
    error = function(e) ""
  )
  if (is.null(err_msg) || err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }
  cli::cli_abort(c("x" = "Native API Failed: {err_msg}"), call = NULL)
}

#' Batch Chat Completion with LM Studio
#'
#' Process a vector of inputs sequentially through LM Studio.
#'
#' @param model Character. The loaded model name.
#' @param inputs Character vector. The prompts to process.
#' @param system_prompt Character. Optional system prompt.
#' @param format Character. Output format: "vector", "list", or "data.frame".
#' @param host Character. Server URL.
#' @param simplify Logical. If TRUE, parses outputs.
#' @param quiet Logical. Whether to suppress the progress bar.
#' @param ... Additional arguments passed to `lms_chat`.
#' @return The return type depends on the \code{format} argument:
#' \itemize{
#'   \item \code{"vector"}: A character vector of responses. This format is only supported if \code{simplify = TRUE} and \code{logprobs = FALSE}.
#'   \item \code{"list"}: A list where each element is the response corresponding to the provided input.
#'   \item \code{"data.frame"}: A data.frame containing \code{input} and \code{output} columns. If \code{logprobs = TRUE}, an additional list-column named \code{logprobs} is included.
#' }
#' @export
lms_chat_batch <- function(
  model,
  inputs,
  system_prompt = NULL,
  format = c("vector", "list", "data.frame"),
  host = "http://localhost:1234",
  simplify = TRUE,
  quiet = FALSE,
  ...
) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }
  format <- match.arg(format)

  if (!is.character(inputs) || length(inputs) == 0) {
    cli::cli_abort(
      "{.arg inputs} must be a non-empty character vector.",
      call = NULL
    )
  }

  args <- list(...)
  has_logprobs <- isTRUE(args$logprobs)
  should_be_quiet <- is_quiet(quiet)

  if (!should_be_quiet) {
    pb <- cli::cli_progress_bar(
      name = "Batch processing",
      total = length(inputs),
      format = "{cli::pb_name} {cli::pb_bar} {cli::pb_percent} | ETA: {cli::pb_eta}"
    )
    on.exit(cli::cli_progress_done(id = pb), add = TRUE)
  }

  results <- lapply(inputs, function(input) {
    res <- lms_chat(
      model = model,
      input = input,
      system_prompt = system_prompt,
      host = host,
      simplify = simplify,
      ...
    )
    if (!should_be_quiet) {
      cli::cli_progress_update(id = pb)
    }
    res
  })

  if (format == "data.frame") {
    if (!isTRUE(simplify)) {
      cli::cli_abort(
        "The {.val data.frame} format requires {.code simplify = TRUE}.",
        call = NULL
      )
    }

    any_logprobs <- any(vapply(
      results,
      inherits,
      logical(1),
      "lms_chat_result"
    ))

    if (any_logprobs) {
      df <- data.frame(
        input = inputs,
        output = vapply(
          results,
          function(x) if (inherits(x, "lms_chat_result")) x$text else x,
          character(1)
        ),
        stringsAsFactors = FALSE
      )
      # Add the logprobs as a list-column
      df$logprobs <- lapply(results, function(x) {
        if (inherits(x, "lms_chat_result")) x$logprobs else NULL
      })
      return(df)
    } else {
      return(data.frame(
        input = inputs,
        output = unlist(results),
        stringsAsFactors = FALSE
      ))
    }
  }

  if (format == "vector") {
    if (!isTRUE(simplify)) {
      cli::cli_warn(
        "The {.val vector} format is not compatible with simplify = FALSE. Returning list."
      )
      return(results)
    }
    if (has_logprobs) {
      cli::cli_warn(
        "The {.val vector} format cannot store logprobs dataframes. Returning list."
      )
      return(results)
    }
    return(unlist(results))
  }

  results
}

#' Create a base request for the LM Studio API
#'
#' @param host Character. The host address of the local server.
#'   Defaults to "http://localhost:1234".
#'
#' @return An httr2 request object.
#'
#' @noRd
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#'
#' req <- lms_client("http://localhost:1234")
#' # req is a base httr2 request object that can be further modified
#' }
lms_client <- function(host = "http://localhost:1234") {
  httr2::request(host) |>
    httr2::req_headers(
      "Content-Type" = "application/json",
      "Accept" = "application/json"
    )
}

#' Unload a model from memory via REST API
#'
#' @param model Character. Unique identifier (\code{instance_id}) of the model
#'   instance to unload.
#' @param host Character. The host address of the local server. Defaults to
#'   "http://localhost:1234".
#' @param ... Additional arguments passed to the API request body.
#'
#' @note If you have loaded multiple instances of the same model using
#'   \code{force = TRUE} in \code{lms_load()}, the server assigns them unique
#'   instance identifiers (e.g., \code{"google/gemma-3-1b"} and
#'   \code{"google/gemma-3-1b:2"}). Passing the base model name to
#'   \code{lms_unload()} will only unload the primary instance. To unload
#'   duplicate instances, you must provide their exact \code{instance_id}, or
#'   use \code{lms_unload_all()} to clear everything.
#'
#' @seealso [LM Studio Unload Model
#'   API](https://lmstudio.ai/docs/developer/rest/unload)
#'
#' @return Invisibly returns a character string representing the unloaded
#'   \code{instance_id} upon success.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#' lms_download("google/gemma-3-1b")
#' lms_load("google/gemma-3-1b")
#'
#' # Unload a single specific model
#' lms_unload("google/gemma-3-1b")
#' }
lms_unload <- function(model, host = "http://localhost:1234", ...) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  # Build body and merge extra args from dots
  body <- utils::modifyList(list(instance_id = model), list(...))

  rlm_progress_step(
    msg = "Unloading model: {.val {model}}...",
    msg_done = "Model {.val {model}} unloaded successfully."
  )

  resp <- lms_client(host) |>
    httr2::req_url_path("api/v1/models/unload") |>
    httr2::req_body_json(body) |>
    httr2::req_error(is_error = \(resp) FALSE) |>
    httr2::req_perform()

  if (httr2::resp_status(resp) == 200) {
    return(invisible(model))
  }

  err_msg <- tryCatch(
    {
      err_json <- httr2::resp_body_json(resp)
      if (!is.null(err_json$error$message)) {
        err_json$error$message
      } else if (!is.null(err_json$error)) {
        err_json$error
      } else {
        httr2::resp_body_string(resp)
      }
    },
    error = function(e) httr2::resp_body_string(resp)
  )

  if (err_msg == "") {
    err_msg <- paste("HTTP Status", httr2::resp_status(resp))
  }

  cli::cli_abort(c("x" = "API Unload Failed: {err_msg}"), call = NULL)
}

#' Unload all models from memory
#'
#' Retrieves a list of all currently loaded models and unloads them one by one.
#'
#' @param host Character. The host address of the local server. Defaults to
#'   "http://localhost:1234".
#' @param ... Additional arguments passed to the API request body for each
#'   unload request.
#'
#' @seealso \code{\link{lms_unload}}
#'
#' @return Invisibly returns a character vector of the \code{instance_id}s that
#'   were successfully unloaded. If no models were currently loaded, it
#'   invisibly returns \code{NULL}.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' lms_server_start()
#' lms_download("google/gemma-3-1b")
#' lms_load("google/gemma-3-1b")
#'
#' # Unload all currently loaded models to clear VRAM
#' lms_unload_all()
#' }
lms_unload_all <- function(host = "http://localhost:1234", ...) {
  if (!is_server_running()) {
    cli::cli_abort(
      "The LM Studio server is not running. Run {.fn lms_server_start} first.",
      call = NULL
    )
  }

  # Fetch currently active models
  active_models <- list_models(
    loaded = TRUE,
    detailed = TRUE,
    quiet = TRUE,
    host = host
  )

  # Check if there is anything to unload
  if (nrow(active_models) == 0 || is.null(active_models$loaded_instances)) {
    rlm_alert_info("No models are currently loaded.")
    return(invisible(NULL))
  }

  # Robustly extract all instance IDs from the nested list column
  loaded_keys <- unlist(lapply(active_models$loaded_instances, function(x) {
    if (is.data.frame(x)) {
      # The API typically uses 'identifier' or 'id' for the specific instance string
      if ("identifier" %in% names(x)) {
        return(x$identifier)
      }
      if ("id" %in% names(x)) {
        return(x$id)
      }
      return(as.character(x[[1]])) # Fallback to the first column
    } else {
      return(as.character(x))
    }
  }))

  # Filter out any potential NAs or empty strings just to be safe
  loaded_keys <- Filter(function(x) !is.na(x) && x != "", loaded_keys)

  if (length(loaded_keys) == 0) {
    rlm_alert_info("No models are currently loaded.")
    return(invisible(NULL))
  }

  rlm_alert_info(
    "Found {length(loaded_keys)} loaded model instance{?s}. Unloading now..."
  )

  # Loop through and unload each specific instance
  for (instance_id in loaded_keys) {
    lms_unload(model = instance_id, host = host, ...)
  }

  rlm_alert_success("All models unloaded successfully.")

  return(invisible(loaded_keys))
}

#' Create a new LM Studio chat result
#'
#' Internal constructor to create a structured object for responses containing
#' log probabilities.
#'
#' @param text Character. The generated text response.
#' @param logprobs Dataframe. The token-level probability data.
#'
#' @return An object of class \code{lms_chat_result}.
#'
#' @keywords internal
new_lms_chat_result <- function(text = character(), logprobs = data.frame()) {
  stopifnot(is.character(text))

  structure(
    list(
      text = text,
      logprobs = logprobs
    ),
    class = "lms_chat_result"
  )
}

#' Validate an LM Studio chat result
#'
#' Internal validator to ensure the integrity of \code{lms_chat_result} objects.
#'
#' @param x An object to validate.
#'
#' @return The validated object.
#'
#' @keywords internal
validate_lms_chat_result <- function(x) {
  if (!is.character(x$text)) {
    cli::cli_abort("{.arg text} must be a character vector.")
  }
  if (!is.null(x$logprobs) && !inherits(x$logprobs, "data.frame")) {
    cli::cli_abort("{.arg logprobs} must be a data.frame or NULL.")
  }
  x
}

#' Print an LM Studio chat result
#'
#' Custom print method for responses that include log probabilities. Displays
#' the text clearly and provides a summary of the metadata.
#'
#' @param x An object of class \code{lms_chat_result}.
#' @param ... Additional arguments passed to print.
#'
#' @return Invisibly returns the input object \code{x}.
#'
#' @keywords internal
#'
#' @export
print.lms_chat_result <- function(x, ...) {
  # Create a clean, themed layout using cli
  cli::cli_div(
    theme = list(rule = list(color = "cyan", "line-type" = "single"))
  )

  cli::cli_rule(left = "{.strong LM Studio Response}")

  # Print the generated text with padding
  cat("\n")
  cli::cli_text("{x$text}")
  cat("\n")

  cli::cli_rule()

  # Provide context about the logprobs if they exist
  if (
    !is.null(x$logprobs) &&
      inherits(x$logprobs, "data.frame") &&
      nrow(x$logprobs) > 0
  ) {
    # Count unique steps by looking at the step_token or step indices
    n_steps <- length(unique(x$logprobs$step_token))

    rlm_alert_info(
      "Includes log probabilities for {.val {n_steps}} token step{?s}. Access the data via {.code x$logprobs}"
    )
  } else {
    cli::cli_alert_warning("Log probability data is empty or missing.")
  }

  # Standard S3 practice: return the object invisibly
  invisible(x)
}

---
title: "Headless Configuration and Usage"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Headless Configuration and Usage}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup}
library(rlmstudio)
lms_installed <- has_lms()

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The `rlmstudio` package provides robust support for running LM Studio in completely headless environments. This is ideal for Linux servers, Docker containers, remote cloud instances, and automated CI/CD pipelines where a visual desktop application is unavailable or inconvenient.

To operate without a GUI, LM Studio relies on a background process called the `llmster` daemon. This vignette will walk you through managing the daemon, starting the local server, and fully automating your local LLM workflows.

## Setup and Installation

If you are setting up a fresh remote server, you can use the package to download and install the LM Studio CLI automatically via the terminal. Run `install_lmstudio(method = "headless")` in your console to execute the automated installation script.

```{r check-status}
# Verify the CLI is available before proceeding
has_lms()
```

## Step-by-Step Guide

### 1. Start the Background Daemon

Unlike the desktop version where opening the app initializes the backend engine, a headless environment requires you to start the engine manually. You must start the `llmster` daemon before attempting to load models or start the API server.

```{r start-daemon, eval=lms_installed}
# Start the headless engine in the background
lms_daemon_start()
```

### 2. Start the Local Server

With the daemon running, you can now spin up the REST API server to accept HTTP requests.

```{r start-server, eval=lms_installed}
# Start the local server on the default port
lms_server_start()
```

### 3. Finding and Managing Models

Because you do not have the GUI's visual search tool, you will need to know the Hugging Face repository or the LM Studio catalog identifier for the model you want to use.

```{r download, eval=lms_installed}
# Download a model using its identifier
job_id <- lms_download("qwen/qwen3-4b-2507")
lms_download_status(job_id)
```

```{r wait, echo=FALSE, eval=lms_installed}
if (job_id != "already_downloaded") {
  repeat {
    res <- lms_download_status(job_id)
    if (res$status %in% c("completed", "failed", "error")) {
      break
    }
    Sys.sleep(3)
  }
  if (res$status != "completed") {
    cli::cli_abort("Model download failed.")
  }
}
```

```{r list, eval=lms_installed}
# View all downloaded models
models <- list_models()

# Filter for unloaded text models
unloaded_llms <- models |>
  subset(type == "llm" & state == "unloaded")
unloaded_llms
```

### 4. Loading Models

Allocate the model to your system's memory (RAM/VRAM) so it is ready for inference.

```{r load, eval=lms_installed}
# Load the model
lms_load("google/gemma-3-1b", flash_attention = TRUE)
```

### 5. Chatting

Interact with the model exactly as you would in a desktop environment.

```{r chat, eval=lms_installed}
response <- lms_chat(
  model = "google/gemma-3-1b",
  input = "Provide just the str_extract() pattern to match all text after the third comma.",
  system_prompt = "You are an expert R programmer familiar with the tidyverse."
)

cat(response)
```

### 6. Teardown and Cleanup

In a headless environment, managing your system resources is critical. When your script finishes, you should explicitly tear down the entire stack to free up memory and stop background processes.

```{r teardown, eval=lms_installed}
# 1. Unload the model from memory
lms_unload("google/gemma-3-1b")

# 2. Stop the API server
lms_server_stop()

# 3. Stop the background daemon
lms_daemon_stop()
```

## Bonus: Pipeline Automation

If you are writing a script that just needs to run a quick job and exit, managing the daemon state manually can be tedious. The `with_lms_daemon()` wrapper handles the setup and guaranteed teardown of the background engine automatically.

```{r with-daemon, eval=lms_installed}
# The daemon will start, the code will run, and the daemon will stop on exit.
results <- with_lms_daemon({
  lms_server_start()
  lms_load("google/gemma-3-1b")

  res <- lms_chat("google/gemma-3-1b", "Is the daemon running?")

  lms_server_stop()
  res
})
```

---
title: "Getting Started with LM Studio in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with LM Studio in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup}
library(rlmstudio)
lms_installed <- has_lms()

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The `rlmstudio` package bridges the gap between R and local Large Language Models by wrapping the LM Studio CLI and its REST API. This vignette covers the **GUI Workflow**, which is best for visual users on desktop environments like macOS, Windows, and Linux desktops.

While the R package provides functions to manage the entire lifecycle of a local LLM, the LM Studio desktop application provides an excellent visual search function for finding new models and exploring advanced configurations beyond what the API can currently do. You can seamlessly mix and match: use the GUI to discover and tweak models, and use R to automate your chatting and data processing.

## Setup and Installation

This package relies on the LM Studio CLI. If you do not have LM Studio installed or need to update your version, the package provides a convenient setup function.

For desktop users, you can run `install_lmstudio(method = "browser")` in your console to open the official download page.

```{r check-status}
# Check if LM Studio is available on this system
has_lms()
```

## Step-by-Step Guide

### 1. Start the Server

You have two options for starting the local server. You can open the LM Studio desktop application, navigate to the Developer or Local Server tab, and click "Start". Alternatively, you can start it directly from R.

```{r start-server, eval=lms_installed}
# Start the local server on the default port
lms_server_start()
#> ✔ LM Studio server started successfully on the default port.
```

### 2. Finding and Managing Models

The LM Studio GUI shines when it comes to discovering models. You can use its built-in search bar to browse Hugging Face, filter by compatibility, and select specific quantizations.

However, if you already know the exact identifier of the model you want, you can download it and manage your inventory directly from R.

```{r download, eval=lms_installed}
# Download a model using its identifier
model <- "google/gemma-3-1b"
job_id <- lms_download(model)
#> ✔ Initiating download for model: "google/gemma-3-1b"... [973ms]
#> ✔ Download job started successfully. Job ID: "job_02c8a1f86e"

lms_download_status(job_id)
#> ── Download Job: "job_02c8a1f86e"
#> Status: completed
#> Progress: 100% (0.72 GB / 0.72 GB)
```

```{r wait, echo=FALSE, eval=lms_installed}
if (job_id != "already_downloaded") {
  repeat {
    res <- lms_download_status(job_id)
    if (res$status %in% c("completed", "failed", "error")) {
      break
    }
    Sys.sleep(3)
  }
  if (res$status != "completed") {
    cli::cli_abort("Model download failed.")
  }
}
```

### 3. Loading Models

Before you can chat with a model, you must load it into system memory. We include the optional `flash_attention = TRUE` argument here, which speeds up processing and reduces memory usage on supported hardware.

```{r load, eval=lms_installed}
# Standard load
lms_load(model, flash_attention = TRUE)
#> ✔ Model "google/gemma-3-1b" loaded and verified. [30.8s]
```

### 4. Chatting

Interact with the model by sending it text prompts. The `lms_chat()` function takes a few key arguments to guide the AI's response:

- `input`: This is your main message or question for the model.

- `system_prompt`: This is an optional set of background instructions. You use it to tell the AI how to behave, what role to play, or how to format its answers (like asking it to act as an expert R programmer).

*Note:* The rlmstudio package currently processes each chat request independently. Every time you run the lms_chat() function, you are starting a brand new conversation. The model will not remember previous messages or context from earlier in your R script.

```{r simple-chat, eval=lms_installed}
response <- lms_chat(
  model = model,
  input = "Say hello!",
  system_prompt = "Answer in rhymes."
)

cat(response)
#> A friendly face, so bright and new,
#> Hello there, it’s waiting for you!
#>
#> Let's chat and have a joyful spree,
#> Hello there, happy as can be!
```

### 5. Teardown

To free up memory and system resources when you are finished, it is best practice to unload your models and stop the local server. Closing the LM Studio GUI will also perform this cleanup if you forget.

```{r teardown, eval=lms_installed}
# Unload the model
lms_unload(model)
#> ✔ Model "google/gemma-3-1b" unloaded successfully. [431ms]

# Stop the server
lms_server_stop()
#> ✔ LM Studio server stopped successfully.
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/daemon.R
\name{lms_daemon_status}
\alias{lms_daemon_status}
\title{Check the global status of LM Studio}
\usage{
lms_daemon_status()
}
\value{
A character vector of the raw CLI output.
}
\description{
Displays the overall status of the LM Studio backend via the CLI, including loaded
models and the server state. This function works regardless of whether
the backend was started via the desktop GUI or the headless daemon.
}
\examples{
\dontrun{
lms_daemon_status()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unload.R
\name{lms_unload_all}
\alias{lms_unload_all}
\title{Unload all models from memory}
\usage{
lms_unload_all(host = "http://localhost:1234", ...)
}
\arguments{
\item{host}{Character. The host address of the local server. Defaults to
"http://localhost:1234".}

\item{...}{Additional arguments passed to the API request body for each
unload request.}
}
\value{
Invisibly returns a character vector of the \code{instance_id}s that
were successfully unloaded. If no models were currently loaded, it
invisibly returns \code{NULL}.
}
\description{
Retrieves a list of all currently loaded models and unloads them one by one.
}
\examples{
\dontrun{
lms_server_start()
lms_download("google/gemma-3-1b")
lms_load("google/gemma-3-1b")

# Unload all currently loaded models to clear VRAM
lms_unload_all()
}
}
\seealso{
\code{\link{lms_unload}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{has_lms}
\alias{has_lms}
\title{Check if LM Studio CLI is installed}
\usage{
has_lms()
}
\value{
A logical scalar: \code{TRUE} if the \code{lms} executable is found
on the system path, and \code{FALSE} otherwise.
}
\description{
Check if LM Studio CLI is installed
}
\examples{
\dontrun{
has_lms()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{check_lms_version}
\alias{check_lms_version}
\title{Check if the installed LM Studio CLI meets the minimum requirement}
\usage{
check_lms_version(min_version = "0.4.0")
}
\arguments{
\item{min_version}{Character string of the required version. Default is
"0.4.0".}
}
\value{
A logical scalar: \code{TRUE} if the LM Studio CLI version meets or
exceeds the specified \code{min_version}, and \code{FALSE} otherwise.
}
\description{
Check if the installed LM Studio CLI meets the minimum requirement
}
\examples{
\dontrun{
check_lms_version("0.4.0")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{lms_download}
\alias{lms_download}
\title{Download a model via REST API}
\usage{
lms_download(model, quantization = NULL, host = "http://localhost:1234", ...)
}
\arguments{
\item{model}{Character. The model to download. Accepts model catalog
identifiers (e.g., "openai/gpt-oss-20b") and exact Hugging Face links.}

\item{quantization}{Character. Optional. Quantization level of the model to
download (e.g., "Q4_K_M"). Only supported for Hugging Face links.}

\item{host}{Character. The host address of the local server. Defaults to
"http://localhost:1234".}

\item{...}{Additional arguments passed to the request.}
}
\value{
A character string containing the download \code{job_id}, or
\code{"already_downloaded"} if already downloaded.
}
\description{
Download a model via REST API
}
\examples{
\dontrun{
lms_server_start()

# Download a model by its HuggingFace identifier
job_id <- lms_download("google/gemma-3-1b")

# Download with a specific quantization level
lms_download("google/gemma-3-1b", quantization = "4bit")
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/rest/download}{LM Studio Download Model API}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{lms_chat}
\alias{lms_chat}
\title{Chat Completion with LM Studio}
\usage{
lms_chat(
  model,
  input,
  system_prompt = NULL,
  host = "http://localhost:1234",
  api_type = c("openresponses", "openai", "native"),
  logprobs = FALSE,
  simplify = TRUE,
  ...
)
}
\arguments{
\item{model}{Character. The name of the loaded model.}

\item{input}{Character. The user prompt to send to the model.}

\item{system_prompt}{Character. An optional system prompt to guide model
behavior.}

\item{host}{Character. The base URL of the LM Studio server. Default is
"http://localhost:1234".}

\item{api_type}{Character. The LM Studio API endpoint to use. Options are
"openresponses" (default), "openai", or "native".}

\item{logprobs}{Logical. Whether to return the log probabilities of the
generated tokens. Default is FALSE.}

\item{simplify}{Logical. If TRUE, extracts the core text response. Default is
TRUE.}

\item{...}{Additional arguments passed to the selected API body.}
}
\value{
Depending on the arguments provided:
\itemize{
\item If \code{simplify = FALSE}, returns a parsed list of the raw JSON response.
\item If \code{simplify = TRUE} and \code{logprobs = FALSE}, returns a single character string containing the model's text response.
\item If \code{simplify = TRUE} and \code{logprobs = TRUE} (and the chosen API type supports it), returns an object of class \code{lms_chat_result} containing both the text and a data.frame of token probabilities.
}
}
\description{
Send a prompt to a locally running LM Studio model. This wrapper
automatically routes your request to the appropriate subfunction based on the
selected API type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unload.R
\name{lms_unload}
\alias{lms_unload}
\title{Unload a model from memory via REST API}
\usage{
lms_unload(model, host = "http://localhost:1234", ...)
}
\arguments{
\item{model}{Character. Unique identifier (\code{instance_id}) of the model
instance to unload.}

\item{host}{Character. The host address of the local server. Defaults to
"http://localhost:1234".}

\item{...}{Additional arguments passed to the API request body.}
}
\value{
Invisibly returns a character string representing the unloaded
\code{instance_id} upon success.
}
\description{
Unload a model from memory via REST API
}
\note{
If you have loaded multiple instances of the same model using
\code{force = TRUE} in \code{lms_load()}, the server assigns them unique
instance identifiers (e.g., \code{"google/gemma-3-1b"} and
\code{"google/gemma-3-1b:2"}). Passing the base model name to
\code{lms_unload()} will only unload the primary instance. To unload
duplicate instances, you must provide their exact \code{instance_id}, or
use \code{lms_unload_all()} to clear everything.
}
\examples{
\dontrun{
lms_server_start()
lms_download("google/gemma-3-1b")
lms_load("google/gemma-3-1b")

# Unload a single specific model
lms_unload("google/gemma-3-1b")
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/rest/unload}{LM Studio Unload Model API}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serve.R
\name{lms_server_start}
\alias{lms_server_start}
\title{Start the LM Studio local server}
\usage{
lms_server_start(port = NULL, cors = FALSE)
}
\arguments{
\item{port}{Integer. Port to run the server on. If not provided, LM Studio
uses the last used port.}

\item{cors}{Logical. Enable CORS support for web application development.
Defaults to FALSE.}
}
\value{
Invisibly returns an integer representing the system exit code
(\code{0} for success).
}
\description{
Launches the LM Studio local server via the CLI, allowing you to interact
with loaded models via HTTP API calls.
}
\examples{
\dontrun{
# Start server on the default port
lms_server_start()

# Start server on a custom port with CORS enabled
lms_server_start(port = 8080, cors = TRUE)
}
}
\seealso{
\href{https://lmstudio.ai/docs/cli/serve/server-start}{LM Studio CLI Server Start Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/daemon.R
\name{lms_daemon_start}
\alias{lms_daemon_start}
\title{Start the LM Studio headless daemon}
\usage{
lms_daemon_start()
}
\value{
Invisibly returns the process object (or 0 if already running).
}
\description{
Launches the \code{llmster} daemon in the background via the CLI. This is required
in headless environments (such as Linux servers) before loading models or
starting the local server.
}
\section{Desktop Users}{
 On desktop operating systems (macOS and Windows),
running this command may actually launch the LM Studio desktop application
to act as the backend engine. If the GUI is already open, this function
will simply detect the active instance and return successfully. While safe
to use, desktop users generally do not need to call this function and can
just open the application manually.
}

\examples{
\dontrun{
lms_daemon_start()
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/core/headless_llmster}{LM Studio Headless Daemon (llmster)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score.R
\name{lms_score_expected}
\alias{lms_score_expected}
\title{Calculate Expected Scores and Uncertainty from Logprobs}
\usage{
lms_score_expected(lp_df, scale = 1:5)
}
\arguments{
\item{lp_df}{A dataframe of logprobs (e.g., \code{x$logprobs}).}

\item{scale}{Numeric vector. The valid labels (e.g., \code{1:5}).}
}
\value{
A named list containing three numeric elements
(\code{expected_value}, \code{weighted_sd}, \code{entropy}) and a
\code{data.frame} named \code{probabilities} with columns \code{label} and
\code{prob}. Returns \code{NULL} if the input dataframe is empty or
invalid.
}
\description{
Takes a logprobs dataframe (from an \code{lms_chat_result}) and calculates
the weighted average score, normalized probabilities, and uncertainty
metrics.
}
\examples{
# Create a sample logprobs dataframe representing a model's generation step
mock_logprobs <- data.frame(
  step_token = rep("4", 3),
  step_logprob = rep(0, 3),
  candidate_token = c("4", "5", "3"),
  candidate_logprob = c(-0.105, -2.302, -3.506),
  stringsAsFactors = FALSE
)

# Calculate the expected score and uncertainty metrics
lms_score_expected(mock_logprobs, scale = 1:5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{install_lmstudio}
\alias{install_lmstudio}
\title{Help the user install or update LM Studio}
\usage{
install_lmstudio(method = c("browser", "headless"))
}
\arguments{
\item{method}{Character. Either "browser" (opens the GUI download page) or
"headless" (installs the \code{llmster} daemon via script).}
}
\value{
Invisibly returns \code{TRUE} upon successful completion. This
function is primarily utilized for its side effects of opening a web
browser or executing system installation commands.
}
\description{
This function provides two methods for setting up LM Studio on your system.
The "browser" method opens the official download page for the LM Studio
desktop application (GUI). The "headless" method runs an automated
installation script to install the \code{llmster} daemon and CLI, which is
suitable for servers, containers, or users who prefer a GUI-less environment.
}
\examples{
\dontrun{
# Open your default web browser to the download page
install_lmstudio(method = "browser")

# Attempt automatic headless installation via the command line
install_lmstudio(method = "headless")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load.R
\name{lms_load}
\alias{lms_load}
\title{Load a model via REST API}
\usage{
lms_load(
  model,
  context_length = NULL,
  eval_batch_size = NULL,
  flash_attention = NULL,
  num_experts = NULL,
  offload_kv_cache_to_gpu = NULL,
  echo_load_config = FALSE,
  force = FALSE,
  host = "http://localhost:1234",
  ...
)
}
\arguments{
\item{model}{Character. Unique identifier for the model to load.}

\item{context_length}{Integer. Maximum number of tokens that the model will
consider.}

\item{eval_batch_size}{Integer. Number of input tokens to process together in
a single batch during evaluation.}

\item{flash_attention}{Logical. Whether to optimize attention computation.}

\item{num_experts}{Integer. Number of experts to use during inference for MoE
models.}

\item{offload_kv_cache_to_gpu}{Logical. Whether KV cache is offloaded to GPU
memory.}

\item{echo_load_config}{Logical. If \code{TRUE}, echoes the final load
configuration in the response.}

\item{force}{Logical. If \code{TRUE}, bypasses the check for currently loaded
models and requests a new instance from the server. Note that this does not
overwrite or replace the existing model; it loads a second concurrent
instance into VRAM. Defaults to \code{FALSE}.}

\item{host}{Character. The host address of the local server. Defaults to
"http://localhost:1234".}

\item{...}{Additional arguments passed to the API request body (useful for
future API parameters).}
}
\value{
Invisibly returns a character string of the loaded model identifier
upon success. If \code{echo_load_config = TRUE}, it instead invisibly
returns a list containing the model's detailed load configuration.
}
\description{
Load a model via REST API
}
\examples{
\dontrun{
lms_server_start()
lms_download("google/gemma-3-1b")

# Load a model with default settings
lms_load("google/gemma-3-1b")

# Load a model with custom context length and flash attention enabled
lms_load("google/gemma-3-1b", context_length = 8192, flash_attention = TRUE)
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/rest/load}{LM Studio Load Model API}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_oop.R
\name{print.lms_chat_result}
\alias{print.lms_chat_result}
\title{Print an LM Studio chat result}
\usage{
\method{print}{lms_chat_result}(x, ...)
}
\arguments{
\item{x}{An object of class \code{lms_chat_result}.}

\item{...}{Additional arguments passed to print.}
}
\value{
Invisibly returns the input object \code{x}.
}
\description{
Custom print method for responses that include log probabilities. Displays
the text clearly and provides a summary of the metadata.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{lms_chat_native}
\alias{lms_chat_native}
\title{Chat Completion via Native API}
\usage{
lms_chat_native(
  model,
  input,
  system_prompt = NULL,
  host = "http://localhost:1234",
  simplify = TRUE,
  ...
)
}
\arguments{
\item{model}{Character. The loaded model name.}

\item{input}{Character. The user prompt.}

\item{system_prompt}{Character. Optional system prompt.}

\item{host}{Character. Server URL.}

\item{simplify}{Logical. If TRUE, parses output to text.}

\item{...}{Additional API arguments.}
}
\value{
If \code{simplify = FALSE}, returns a list representing the raw JSON
response. If \code{simplify = TRUE}, returns a character string containing
the model's text output.
}
\description{
Direct interface to LM Studio's v1 Native endpoint. Optimized for stateful chats and hardware control.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{lms_chat_openai}
\alias{lms_chat_openai}
\title{Chat Completion via OpenAI Compatibility API}
\usage{
lms_chat_openai(
  model,
  messages,
  host = "http://localhost:1234",
  logprobs = FALSE,
  simplify = TRUE,
  ...
)
}
\arguments{
\item{model}{Character. The loaded model name.}

\item{messages}{List. A structured list of role and content pairs.}

\item{host}{Character. Server URL.}

\item{logprobs}{Logical. Whether to request logprobs (currently stubbed by LM
Studio).}

\item{simplify}{Logical. If TRUE, parses output to text.}

\item{...}{Additional API arguments.}
}
\value{
If \code{simplify = FALSE}, returns a list representing the raw JSON
response. Otherwise, returns a character string containing the generated
text. If \code{logprobs = TRUE}, it returns an \code{lms_chat_result}
object with the log probabilities populated as \code{NULL} since they are
currently stubbed in the LM Studio OpenAI endpoint.
}
\description{
Direct interface to LM Studio's OpenAI-compatible endpoint. Uses the messages
array format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/daemon.R
\name{lms_daemon_stop}
\alias{lms_daemon_stop}
\title{Stop the LM Studio headless daemon}
\usage{
lms_daemon_stop(force = FALSE)
}
\arguments{
\item{force}{Logical. If \code{TRUE}, attempts to stop the local server before
shutting down the daemon. The daemon cannot be stopped while the server
is actively running. Defaults to \code{FALSE}.}
}
\value{
Invisibly returns the system exit code (0 for success).
}
\description{
Stops the \code{llmster} daemon via the CLI. Use this to clean up system resources when
you are completely finished using LM Studio in headless mode.
}
\section{Desktop Users}{

If the daemon is currently being managed by the LM Studio desktop
application, this function will fail. The CLI intentionally prevents
programmatic shutdowns of the GUI to avoid disrupting visual sessions.
In this scenario, you must close the desktop application manually.
}

\examples{
\dontrun{
lms_daemon_stop(force = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{lms_path}
\alias{lms_path}
\title{Get the absolute path to the LMS executable}
\usage{
lms_path()
}
\value{
A character string specifying the absolute file path to the LM Studio
executable (\code{lms}) on the user's system.
}
\description{
Locates the LM Studio CLI (\code{lms}) on your system. It checks the
\code{RLMSTUDIO_LMS_PATH} environment variable first, then the system \code{PATH}, and
finally common installation directories.
}
\examples{
\dontrun{
lms_path()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{lms_chat_batch}
\alias{lms_chat_batch}
\title{Batch Chat Completion with LM Studio}
\usage{
lms_chat_batch(
  model,
  inputs,
  system_prompt = NULL,
  format = c("vector", "list", "data.frame"),
  host = "http://localhost:1234",
  simplify = TRUE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{model}{Character. The loaded model name.}

\item{inputs}{Character vector. The prompts to process.}

\item{system_prompt}{Character. Optional system prompt.}

\item{format}{Character. Output format: "vector", "list", or "data.frame".}

\item{host}{Character. Server URL.}

\item{simplify}{Logical. If TRUE, parses outputs.}

\item{quiet}{Logical. Whether to suppress the progress bar.}

\item{...}{Additional arguments passed to \code{lms_chat}.}
}
\value{
The return type depends on the \code{format} argument:
\itemize{
\item \code{"vector"}: A character vector of responses. This format is only supported if \code{simplify = TRUE} and \code{logprobs = FALSE}.
\item \code{"list"}: A list where each element is the response corresponding to the provided input.
\item \code{"data.frame"}: A data.frame containing \code{input} and \code{output} columns. If \code{logprobs = TRUE}, an additional list-column named \code{logprobs} is included.
}
}
\description{
Process a vector of inputs sequentially through LM Studio.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/daemon.R
\name{with_lms_daemon}
\alias{with_lms_daemon}
\title{Run code with the LM Studio daemon active}
\usage{
with_lms_daemon(code)
}
\arguments{
\item{code}{An R expression to execute while the daemon is running.}
}
\value{
The result of the evaluated code.
}
\description{
Temporarily starts the LM Studio headless daemon, executes the provided
R expression, and then gracefully shuts the daemon and any active servers
down. This is ideal for automated scripts and pipelines.
}
\section{Desktop Users}{

Be cautious using this wrapper if you already have the LM Studio GUI open.
While the setup phase (\code{lms_daemon_start}) will succeed, the teardown phase
(\code{lms_daemon_stop}) will fail because the CLI prevents programmatic shutdowns
of the graphical interface. This wrapper is best reserved for strictly
headless environments or fully automated scripts.
}

\examples{
\dontrun{
result <- with_lms_daemon({
  lms_load("llama-3.1-8b")
  lms_chat("llama-3.1-8b", input = "Hello world!")
})
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serve.R
\name{lms_server_status}
\alias{lms_server_status}
\title{Check the status of the LM Studio server}
\usage{
lms_server_status(
  json = FALSE,
  verbose = FALSE,
  quiet = FALSE,
  log_level = NULL
)
}
\arguments{
\item{json}{Logical. Output the status in machine-readable JSON format.}

\item{verbose}{Logical. Enable detailed logging output.}

\item{quiet}{Logical. Suppress all logging output.}

\item{log_level}{Character. The level of logging to use (e.g., "info",
"debug").}
}
\value{
By default, returns a character vector containing the raw CLI output.
If \code{json = TRUE} and the \code{jsonlite} package is available, it
returns a parsed list or \code{data.frame} of the status configuration.
}
\description{
Displays the current status of the LM Studio local server via the CLI,
including whether it is running and its configuration.
}
\details{
You can only use one logging control flag at a time (\code{verbose},
\code{quiet}, or \code{log_level}).
}
\examples{
\dontrun{
lms_server_start()

# Get basic status string
lms_server_status()

# Get status as a parsed JSON data frame
lms_server_status(json = TRUE)
}
}
\seealso{
\href{https://lmstudio.ai/docs/cli/serve/server-status}{LM Studio CLI Server Status Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{print.lms_download_status}
\alias{print.lms_download_status}
\title{Print method for LM Studio download status}
\usage{
\method{print}{lms_download_status}(x, ...)
}
\arguments{
\item{x}{An object of class \code{lms_download_status}.}

\item{...}{Additional arguments passed to print.}
}
\value{
Invisibly returns the input object \code{x}.
}
\description{
Print method for LM Studio download status
}
\examples{
\dontrun{
lms_server_start()

job_id <- lms_download("google/gemma-3-1b")
status <- lms_download_status(job_id)
print(status)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list.R
\name{list_models}
\alias{list_models}
\title{List available models}
\usage{
list_models(
  loaded = FALSE,
  type = c("llm", "embedding"),
  detailed = FALSE,
  quiet = FALSE,
  host = "http://localhost:1234"
)
}
\arguments{
\item{loaded}{Logical. If \code{TRUE}, returns only currently loaded models.
Defaults to \code{FALSE}.}

\item{type}{Character vector. The types of models to include. Defaults to
\code{c("llm", "embedding")}.}

\item{detailed}{Logical. Show all information about each model. Defaults to
\code{FALSE}.}

\item{quiet}{Logical. If \code{TRUE}, suppresses informative console
messages. Defaults to \code{FALSE}.}

\item{host}{Character. The host address of the local server.}
}
\value{
A \code{data.frame} containing information about the available
models. By default, it includes columns for \code{state}, \code{type},
\code{display_name}, \code{key}, \code{architecture}, and \code{size_gb}.
If \code{detailed = TRUE}, it returns a comprehensive \code{data.frame}
including all raw metadata columns provided by the API. Returns an empty
\code{data.frame} if no models match the criteria.
}
\description{
Retrieves a list of models available on your system via the LM Studio REST
API.
}
\examples{
\dontrun{
lms_server_start()
lms_download("google/gemma-3-1b")
lms_load("google/gemma-3-1b")

# List all downloaded models
list_models()

# List only currently loaded models
list_models(loaded = TRUE)

# Get detailed information about loaded text models
list_models(loaded = TRUE, type = "llm", detailed = TRUE)
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/rest/list}{LM Studio List Models API}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_oop.R
\name{new_lms_chat_result}
\alias{new_lms_chat_result}
\title{Create a new LM Studio chat result}
\usage{
new_lms_chat_result(text = character(), logprobs = data.frame())
}
\arguments{
\item{text}{Character. The generated text response.}

\item{logprobs}{Dataframe. The token-level probability data.}
}
\value{
An object of class \code{lms_chat_result}.
}
\description{
Internal constructor to create a structured object for responses containing
log probabilities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{lms_download_status}
\alias{lms_download_status}
\title{Get the status of a download job}
\usage{
lms_download_status(job_id, host = "http://localhost:1234")
}
\arguments{
\item{job_id}{Character. The unique identifier for the download job.}

\item{host}{Character. The host address of the local server. Defaults to
"http://localhost:1234".}
}
\value{
An object of class \code{lms_download_status} containing the download
status.
}
\description{
Get the status of a download job
}
\examples{
\dontrun{
lms_server_start()

job_id <- lms_download("google/gemma-3-1b")
status <- lms_download_status(job_id)
print(status)
}
}
\seealso{
\href{https://lmstudio.ai/docs/developer/rest/download-status}{LM Studio Download Status API}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rlmstudio-package.R
\docType{package}
\name{rlmstudio-package}
\alias{rlmstudio}
\alias{rlmstudio-package}
\title{rlmstudio: Access and Control LM Studio}
\description{
A community-maintained 'R' wrapper for the 'LM Studio' command line interface and API. Provides functions to manage the local daemon and server, download and load models, and interact with Large Language Models (LLMs).
}
\seealso{
Useful links:
\itemize{
  \item \url{https://jmgirard.github.io/rlmstudio/}
  \item \url{https://github.com/jmgirard/rlmstudio}
  \item Report bugs at \url{https://github.com/jmgirard/rlmstudio/issues}
}

}
\author{
\strong{Maintainer}: Jeffrey Girard \email{me@jmgirard.com} (\href{https://orcid.org/0000-0002-7359-3746}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Jeffrey Girard \email{me@jmgirard.com} (\href{https://orcid.org/0000-0002-7359-3746}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_oop.R
\name{validate_lms_chat_result}
\alias{validate_lms_chat_result}
\title{Validate an LM Studio chat result}
\usage{
validate_lms_chat_result(x)
}
\arguments{
\item{x}{An object to validate.}
}
\value{
The validated object.
}
\description{
Internal validator to ensure the integrity of \code{lms_chat_result} objects.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serve.R
\name{lms_server_stop}
\alias{lms_server_stop}
\title{Stop the LM Studio local server}
\usage{
lms_server_stop()
}
\value{
Invisibly returns an integer representing the system exit code
(\code{0} for success).
}
\description{
Stops the currently running LM Studio local server via the CLI.
}
\examples{
\dontrun{
lms_server_start()
lms_server_stop()
}
}
\seealso{
\href{https://lmstudio.ai/docs/cli/serve/server-stop}{LM Studio CLI Server Stop Documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{lms_chat_openresponses}
\alias{lms_chat_openresponses}
\title{Chat Completion via OpenResponses API}
\usage{
lms_chat_openresponses(
  model,
  input,
  instructions = NULL,
  host = "http://localhost:1234",
  logprobs = FALSE,
  simplify = TRUE,
  ...
)
}
\arguments{
\item{model}{Character. The loaded model name.}

\item{input}{Character. The user prompt.}

\item{instructions}{Character. Optional system instructions.}

\item{host}{Character. Server URL.}

\item{logprobs}{Logical. Whether to return token probabilities.}

\item{simplify}{Logical. If TRUE, parses output to text and dataframe. If
FALSE, returns raw list.}

\item{...}{Additional API arguments (e.g., top_logprobs, temperature).}
}
\value{
If \code{simplify = FALSE}, returns a list representing the raw JSON
response. Otherwise, returns a character string containing the generated
text. If \code{logprobs = TRUE}, returns an object of class
\code{lms_chat_result} incorporating both the text and probability data.
}
\description{
Direct interface to LM Studio's OpenResponses endpoint. Supports logprobs and
custom instructions.
}
